package display;

import Jama.EigenvalueDecomposition;
import Jama.Matrix;
import mPack.Estimator;
import mPack.Motion;
import mPack.SFM_Model;
import mPack.SenseData;
import mPack.Truth;
import utilPack.Util;

public class Reporter {
	
	private Truth truth;

	private SenseData sense;
	
	private Estimator estimator;

	private SFM_Model model;
	
	private Motion motion;
	
	//-----------------------------------------------------------------------
	//
	//-----------------------------------------------------------------------

	
	public Reporter(Truth truth, SenseData sense,Estimator estimator,Motion motion)
	{
		this.truth = truth;
		
		this.sense = sense;
		
		this.estimator = estimator;
		
		this.motion  = motion;
		
	}
	
	public void setModel(SFM_Model model) {this.model = model;}
	
	//--------------------------------------------------------------------------------------------------------------------
	//                                      writing outputs and testing
	//--------------------------------------------------------------------------------------------------------------------
	
	/**
	 *  test second derivative matrices for zero steps
	 *  (there should be one eigenvalues near zero for both bee and bat)
	 */
	public void testSecondDerivs()
	{
		double jitter = 0.1; 
		Util.message("\nTesting second derivative matrix with jitter " + jitter);
		int randoms = 3;
		for (int r = 0; r< randoms; r++)
		{
			for (int obj = 0; obj < truth.allPoints; obj++)
			{
				// add random jitter to the true object position
				double[] objPos = model.makeVec(truth.objectLocation[truth.currentStep][obj]);
				String line = ("\ntest " + r + " of object "  + obj + " at position ");
				for (int dim = 0; dim < 3; dim++)
				{
					line = line  + "\t" + Double.valueOf(objPos[dim]).toString();							
					objPos[dim] = objPos[dim] + jitter*(Math.random() - 0.5);
				}
				Util.message(line);

				// make the second derivative matrix at the semi-random position estimate, for one piece of sense data = line of sight or range
				int calcMethod = estimator.START_BAYES;
				Matrix secondDeriv = estimator.secondDeriv(obj,objPos,calcMethod);

				EigenvalueDecomposition eComp = secondDeriv.eig();
				Matrix eVals = eComp.getD();
				line = "Eigenvalues for object " + obj + "\t";
				
				for (int dim = 0; dim < 3; dim++)
				{
					double eVal = eVals.get(dim,dim);
					line = line + "\t" + Double.valueOf(eVal).toString();
				}
				Util.message(line);
				
			}
		}
		
	}

	
	
	
	/**
	 * 
	 * 
	 * @param obj index of the object
	 * @param estimatePos estimated position of the object; a 3*1 Matrix
	 * @param eVecs eigenvectors of the second derivative matrix
	 * @param eVals eigenvalues of the second derivative matrix
	 * 
	 * @return contribution of one object to chiSquared
	 */
	public double dChiSquared(int obj,Matrix estimatePos, Matrix eVecs, Matrix eVals, int calcMethod)
	{
		double dChi2 = 0.0;
		
		// find the true position of the object
		Matrix objPos = truth.objectLocation[truth.currentStep][obj];
		
		
		// displacement of the true position from the estimate
		Matrix error = objPos.minus(estimatePos);
		
		
		//re-compute the second derivative matrix
		Matrix deriv2 = estimator.secondDeriv(obj, model.makeVec(estimatePos),calcMethod);		
		//  compute 2.0 times what the likelihood gap would be if it were pure quadratic
		double quadGap  = 2.0*error.transpose().times(deriv2).times(error).get(0, 0);
		
		// eigenvector check;loop over eigenvectors and eigenvalues
		for (int eiv = 0; eiv < 3; eiv++)
		{
			// the component of the error along the eigenvector
			double dotp = 0.0;

			// second index denotes which eigenvector
			for (int dim = 0; dim < 3; dim ++)
				dotp = dotp + error.get(dim,0)*eVecs.get(dim, eiv);
			
			// the standard deviation is 1.0/sqrt(2.0*eigenvalue)
			double standardDev = 1.0/Math.sqrt(2.0*eVals.get(eiv, eiv));
			
			// the normalised error is the error divided by the standard deviation
			double normError = dotp/standardDev;
			
			// increment chiSquared by normalised error squared
			dChi2 = dChi2 + normError*normError;
			
		}
		
		// this check would only work if negative log likelihood was pure quadratic
		double likely1 = estimator.negLogLikelihood(obj, model.makeVec(objPos),null,calcMethod);
		double likely2 = estimator.negLogLikelihood(obj, model.makeVec(estimatePos),null, calcMethod);
		double dCheck = likely1 - likely2;

		double ratio = dChi2/quadGap;
		// Util.message("\nChi squared check for object " + obj + "\t" + Util.round(quadGap,4) + "\t" + Util.round(dChi2,4) + "\t" + Util.round(ratio,4));
		
		
		return dChi2;
	}
	
	/**
	 * 
	 * @param obj
	 * @param estimatePos
	 */
	public void interpolateCheck(int obj,Matrix estimatePos, int calcMethod)
	{
		// find the true position of the object
		Matrix objPos = truth.objectLocation[truth.currentStep][obj];
				
		// displacement of the true position from the estimate
		Matrix fullDisp = objPos.minus(estimatePos);
		
		// negative log likelihood at the fitted minimum
		double minLog = estimator.negLogLikelihood(obj, model.makeVec(estimatePos), null,calcMethod);
		
		// second derivative matrix at the minimum
		Matrix deriv2 = estimator.secondDeriv(obj, model.makeVec(estimatePos),calcMethod);		
		
		// number of points to interpolate between the estimate and the true object position; eg from 1/12 to 12/12 displacement
		int points = 10;

		String header  = "Index\t\t";
		String distRow = "Distance\t";
		String realGaps = "Real gap\t";
		String quadGaps = "Quad gap\t";
		String compare = "Ratio\t\t";
		
		for (int p = 1; p < points + 1; p++)
		{
			header = header + p + "\t";
			double ratio = ((double) p)/((double) points);
			
			//scaled displacement 
			Matrix trialDisp = fullDisp.times(ratio);
			double dist = Math.sqrt(trialDisp.transpose().times(trialDisp).get(0, 0));
			distRow = distRow + Util.round(dist, 4) + "\t";

			//displaced position
			Matrix trialPos = estimatePos.plus(trialDisp);

			// true negative log likelihood at displaced position, and likelihood gap
			double negLog = estimator.negLogLikelihood(obj, model.makeVec(trialPos), null, calcMethod);
			double realGap = negLog - minLog;
			realGaps = realGaps + Util.round(realGap, 4) + "\t";
			
			// quadratic negative log likelihood gap at displaced position
			double quadGap  = trialDisp.transpose().times(deriv2).times(trialDisp).get(0, 0);
			quadGaps = quadGaps + Util.round(quadGap, 4) + "\t";
			
			// comparison
			double comparison = quadGap/realGap;
			compare = compare + Util.round(comparison, 4) + "\t";
		}
		
		Util.message("Interpolation check for object "+ obj);
		Util.message(header);
		Util.message(distRow);
		Util.message(realGaps);
		Util.message(quadGaps);
		Util.message(compare + "\n");
		
		

	}

	
	/**
	 * 
	 * @param step
	 */
	public void writeDistances(int step, boolean doMeans)
	{
		String header = "\nStep " + step + " distances with memory noise = 10**-(" + estimator.mem_noise + ")"
				+  "\nObj\tColour\t\tG\tM\tM/G\t\tT-G\t(T-G)/G\t\tN-T\t(N-T)/T";
		Util.message(header);

		for (int obj = 0; obj < truth.allPoints; obj++)
		{
			String line = Integer.valueOf(obj).toString() + "\t" + DrawArea.colour(obj) + "\t";
			Matrix gaussian = estimator.locationEstimate[estimator.FULLBAYES][step][obj];
			Matrix meanEst = estimator.locationEstimate[estimator.MEAN][step][obj];
			Matrix trackEst = estimator.locationEstimate[estimator.TRACK][step][obj];
			Matrix nTrackEst = estimator.locationEstimate[estimator.NOISYTRACK][step][obj];
			Matrix realObj = truth.objectLocation[step][obj];
			
			// columns for Gaussian (g) and Mean estimate (m) distances from the real object position, and their ratio m/g
			String mCols = "\t-\t-\t-\t";
			// columns for tracked estimate (T) distances from the Gaussian object position (T-G), and the ratio (T-G)/G
			String tCols = "\t-\t-\t";
			// columns for noisy tracked estimate (N) distances from the tracked object position (T), and the ratio (N-T)/T
			String nCols = "\t-\t-\t";

			if (estimator.estimated[estimator.FULLBAYES][step][obj])
			{
				// distance of Gaussian from real object
				Matrix gDiff = gaussian.minus(realObj);
				double gDist = Math.sqrt(gDiff.transpose().times(gDiff).get(0, 0));

				// distance of Mean from real object
				double mDist = 0.0;
				if (doMeans)
				{
					Matrix mDiff = meanEst.minus(realObj);
					mDist = Math.sqrt(mDiff.transpose().times(mDiff).get(0, 0));
				}

				double mRatio = mDist/gDist;
				
				mCols = "\t" +  Util.round(gDist, 4) + "\t" +  Util.round(mDist, 4) + "\t" +  Util.round(mRatio, 4) + "\t";

				if (estimator.estimated[estimator.TRACK][step][obj])
				{
					// distance of Tracked estimate from Gaussian estimate
					Matrix tDiff = trackEst.minus(gaussian);
					double tDist = Math.sqrt(tDiff.transpose().times(tDiff).get(0, 0));

					double tRatio = tDist/gDist;
					
					tCols = "\t" + Util.round(tDist, 4) + "\t"  +  Util.round(tRatio, 4) + "\t";

					if (estimator.estimated[estimator.NOISYTRACK][step][obj])
					{
						// distance of noisy Tracked estimate from tracked estimate
						Matrix ntDiff = nTrackEst.minus(trackEst);
						double ntDist = Math.sqrt(ntDiff.transpose().times(ntDiff).get(0, 0));

						double ntRatio = ntDist/tDist;
						
						nCols = "\t" + Util.round(ntDist, 4) + "\t"  +  Util.round(ntRatio, 4) + "\t";
					}
				}
			}
			
			Util.message(line + mCols + tCols + nCols);
					
		}

		
	}

	/**
	 * 
	 * @param step
	 */
	public void writeErrorBars(int step)
	{
		String header = "\nStep " + step + " Error bars\nObj\tColour\t\tG1\tG2\tG3\t\tT1\tT2\tT3\t\tmaxT/maxG";
		Util.message(header);

		for (int obj = 0; obj < truth.allPoints; obj++)
		{
			String line = Integer.valueOf(obj).toString() + "\t" + DrawArea.colour(obj) + "\t";
			Matrix gaussianErrors = estimator.locErrorBars[estimator.FULLBAYES][step][obj];
			Matrix trackedErrors = estimator.locErrorBars[estimator.TRACK][step][obj];
			
			// columns for Gaussian error bars
			String mCols = "\t-\t-\t-\t";
			// columns for tracked error bars, and ratio of maximum error bar
			String tCols = "\t-\t-\t-\t\t-\t";
			
			if (estimator.estimated[estimator.FULLBAYES][step][obj])
			{
				mCols = "\t";
				Matrix[] eigenG = Util.eigenCheck(gaussianErrors,false);
				double maxG = 0.0;
				for (int ev = 0; ev < 3; ev++)
				{
					double g  = eigenG[0].get(ev, ev);
					if (g > maxG) maxG = g;
					mCols = mCols + Util.round(g, 4) + "\t";
				}
				
				if (estimator.estimated[estimator.TRACK][step][obj])
				{
					tCols = "\t";
					Matrix[] eigenT = Util.eigenCheck(trackedErrors,false);
					double maxT = 0.0;
					for (int ev = 0; ev < 3; ev++)
					{
						double t = eigenT[0].get(ev, ev);
						if (t > maxT) maxT = t;
						tCols = tCols + Util.round(t, 4) + "\t";
					}
					tCols = tCols + "\t" + Util.round(maxT/maxG, 4); 
				}
			}
			
			// for debugging - these matrices should be equal, if track2 is defined
			Matrix gauss2 = estimator.negLog2Deriv [estimator.FULLBAYES][step][obj];
			Matrix prevGauss2 = estimator.negLog2Deriv [estimator.FULLBAYES][step - 1][obj];
			Matrix track2 = estimator.negLog2Deriv[estimator.TRACK][step][obj];
			
			Util.message(line + mCols + tCols);
					
		}

		
	}
	
	 /** 
	 * @param step
	 */
	public void realityCheck(int step)
	{
		String header = "\nStep " + step + " Neg log l'hoods\tGFit\tGReal\tDiff\t\tTFit\tTreal\tDiff";
		Util.message(header);

		// sums over all objects
		double sumGE  = 0.0;
		double sumGR  = 0.0;
		double sumTE  = 0.0;
		double sumTR  = 0.0;
		
		for (int obj = 0; obj < truth.allPoints; obj++)
		{
			String line = Integer.valueOf(obj).toString() + "\t" + DrawArea.colour(obj) + "\t";
			
			// columns for Gaussian negative log likelihoods
			String mCols = "\t-\t-\t-";
			// columns for tracked negative log likelihoods
			String tCols = "\t\t-\t-\t-";

			Matrix gaussian = estimator.locationEstimate[estimator.FULLBAYES][step][obj];
			Matrix trackEst = estimator.locationEstimate[estimator.TRACK][step][obj];
			Matrix realObj = truth.objectLocation[step][obj];
			

			if (estimator.estimated[estimator.FULLBAYES][step][obj])
			{				
				int calcMethod = estimator.RUN_BAYES;

				double estLik = estimator.negLogLikelihood(obj, model.makeVec(gaussian), null, calcMethod);
				double realLik = estimator.negLogLikelihood(obj, model.makeVec(realObj), null, calcMethod);
				double likDiff = realLik - estLik;
				
				sumGE = sumGE + estLik;
				sumGR = sumGR + realLik;
				
				mCols = "\t" + Util.round(estLik, 4) + "\t" + Util.round(realLik, 4)+ "\t" + Util.round(likDiff, 4);
				
				if (estimator.estimated[estimator.TRACK][step][obj])
				{
					calcMethod = estimator.START_TRACK;
					if (estimator.estimated[estimator.TRACK][step-1][obj])  calcMethod = estimator.RUN_TRACK;
					
					double estTrack = estimator.negLogLikelihood(obj, model.makeVec(trackEst), null, calcMethod);
					double realTrack = estimator.negLogLikelihood(obj, model.makeVec(realObj), null, calcMethod);
					double trackDiff = realTrack - estTrack;
					
					sumTE = sumTE + estTrack;
					sumTR = sumTR + realTrack;
					
					tCols = "\t\t" + Util.round(estTrack, 4) + "\t" + Util.round(realTrack, 4) + "\t" + Util.round(trackDiff, 4);
				}
			}
						
			Util.message(line + mCols + tCols);					
		}
		
		double diffG = sumGR - sumGE;
		double diffT = sumTR - sumTE;
		Util.message("\nSum: " + "\t\t\t" + Util.round(sumGE, 4) + "\t" + Util.round(sumGR, 4) + "\t" + Util.round(diffG, 4)
		+ "\t\t" + Util.round(sumTE, 4) + "\t" + Util.round(sumTR, 4) + "\t" + Util.round(diffT, 4));

		
	}

}
