package mPack;

import Jama.Matrix;
import utilPack.Util;

/**
 * a balanced binary tree of estimates of point positions,
 * putting nearby points close in the tree
 * 
 *  */

public class BinaryTree {
	
	private Estimator estimator;
	
	// non-leaf trees with two child trees
	public BinaryTree left;
	public BinaryTree right;
	
	// the parent binary tree
	public BinaryTree parent = null;

	// estimated centre of gravity of all nodes
	public Matrix centre;
	
	// position of the animal at this step
	public Matrix animalPos;
	
	// range from the centre of gravity to the animal
	public double animalRange;
	
	// relative storage error to be added to leaf point locations in the tree
	public Matrix relativeStorageError;
	
	// absolute storage errors
	public Matrix absoluteStorageError;
	
	// spatial displacement between left and right subtrees
	public Matrix disp;
	
	// for leaf trees (one point) the index of the point; or -1 if this is  not a leaf
	public int obj;
		
	// half range between centres of gravity of two subtrees, with a maximum half-range of 1.0
	public double splitRange() 
	{
		double range = 0.0;
		if (!isLeaf()) 
		{
			range =  0.5*Math.sqrt(disp.transpose().times(disp).get(0, 0));
			// large range cutoff (for when estimation errors are large)
			double cutoff = 1.0;
			if (range > cutoff) range = cutoff;
		}
		return range;
	}
	
	// total number of leaf nodes (points)
	public int leafCount()
	{
		if (!isLeaf()) return (left.leafCount() + right.leafCount());
		return 1;
	}
	
	// true for a leaf point
	public  boolean isLeaf() {return (obj > -1);}
	
	//---------------------------------------------------------------------------
	//
	//---------------------------------------------------------------------------
	
	
	/**
	 * leaf trees = one point
	 * @param estimator
	 * @param animalPos position of the animal
	 * @param pointPos estimated position of the point
	 * @param obj index of the point
	 */
	public BinaryTree(Estimator estimator,Matrix animalPos,Matrix pointPos,int obj)
	{
		this.estimator = estimator;
		centre = pointPos;
		this.obj = obj;
		this.animalPos = animalPos;
		disp = Matrix.identity(3, 1).times(0.0);
		
		setAnimalRange();
		setZeroErrors();
	}
	
	/**
	 * non-leaf trees - two child trees
	 * @param left
	 * @param right
	 */
	public BinaryTree(Estimator estimator, BinaryTree left, BinaryTree right)
	{
		this.estimator = estimator;
		this.left = left;
		this.right = right;
		left.parent = this;
		right.parent = this;
		this.animalPos = left.animalPos;
		
		obj = -1;
		
		centre = left.centre.plus(right.centre).times(0.5);
		disp = left.centre.minus(right.centre);
		
		setAnimalRange();
		setZeroErrors();
	}
	
	/**
	 * 
	 */
	private void setAnimalRange()
	{
		Matrix aDisp = centre.minus(animalPos);
		animalRange =  Math.sqrt(aDisp.transpose().times(aDisp).get(0, 0));
	}

	
	/**
	 * set errors zero
	 */
	public void setZeroErrors()
	{
		relativeStorageError = Matrix.identity(3, 1).times(0.0);
		absoluteStorageError = Matrix.identity(3, 1).times(0.0);
		
		if (left != null) left.setZeroErrors();
		if (right != null) right.setZeroErrors();
	}
	
	//---------------------------------------------------------------------------
	//
	//---------------------------------------------------------------------------
	
	
	/**
	 * adding a position error to an estimated position
	 * @param error
	public void addError(Matrix error, double scale)
	{
		centre.plusEquals(error.times(scale));
	}
	 */
	
	
	/**
	 * 
	 * @param relative
	 * @return variance sum of all leaf points, if the base standard deviation is 1
	 */
	public double variance(boolean relative)
	{
		double var = 0.0;
		if (relative)
		{
			double split = 1.0; // effective split distance for the top node, with no parent
			if (parent != null) split = parent.splitRange(); // split distance for all other nodes

			// the variance of this node contributes to all leaf nodes below it; so it has weight leafCount().
			// the relative error vector is multiplied by (animalRange*split); so that is squared in the variance.
			var = (animalRange*split)*(animalRange*split)*leafCount();

			// leaf nodes do not contribute to variance
			if (!isLeaf())
			{
				var = var + left.variance(true) + right.variance(true);
			}
		}
		
		else if (!relative)
		{
			// absolute error vector is multiplied by animalRange; square it for the variance
			if (isLeaf()) var = animalRange*animalRange;
			
			// recursive addition down to leaves
			else if (!isLeaf()) var = left.variance(false) + right.variance(false);
		}

		
		return var;
	}
	
	/**
	 * 
	 * @param scale compensation scaling factor, so that squared relative errors summed over all points (variances)
	 * are equal to absolute variances
	 * @param relative rue for relative errors
	 */
	public void findStorageError(double scale, int step, boolean relative)
	{
		if (relative) findRelativeStorageError(scale,step);
		else findAbsoluteStorageError(step);
	}
	
	/**
	 * recursively set storage errors down the tree
	 * @param scale compensation scaling factor, so that squared relative errors summed over all points
	 * are equal to squared absolute errors summed over points, on average
	 */
	public void findRelativeStorageError(double scale, int step)
	{
		// error for the top node assumes range split = 1.0
		if (parent == null) 
		{
			double rangeFactor = scale;
			// error for this (top) node - depends on its location 
			relativeStorageError = estimator.fullMemoryNoise(step,centre).times(rangeFactor); 
		}
		
		// non-leaves set leaves downwards; do nothing at leaves
		if (!isLeaf())
		{
			// compute absolute memory noise for two child nodes, 
			// depending on this node's centre location and the animal's location at this step
			Matrix rawChildNoise = estimator.fullMemoryNoise(step,centre); 
			
			/* left and right subtrees inherit the storage error of this node, 
			 * with an additional error proportional to the distance between them */
			double rangeFactor = scale*splitRange();

			// add the new error vector with opposite signs for left and right
			left.relativeStorageError =  relativeStorageError.plus(rawChildNoise.times(rangeFactor));
			right.relativeStorageError =  relativeStorageError.minus(rawChildNoise.times(rangeFactor));
			
			// carry on down to leaves
			left.findRelativeStorageError(scale, step);
			right.findRelativeStorageError(scale, step);
		}
	}
	
	/**
	 * set absolute storage errors on leaf objects.
	 * with no scaling and no dependence on ranges between internal nodes of the tree
	 * @param step
	 */
	public void findAbsoluteStorageError(int step)
	{
		if (!isLeaf())
		{
			// simple recursion through non-leaf nodes
			left.findAbsoluteStorageError(step);
			right.findAbsoluteStorageError(step);
		}
		
		// leaf objects
		else if (isLeaf())
		{
			// compute absolute memory noise, depending on the object location and the animal's location at this step
			absoluteStorageError = estimator.fullMemoryNoise(step,centre); 		
		}
		
	}

	
	/**
	 * 
	 * @return the sum of squared distance errors over all leaves
	 */
	public double leafErrorSquaredSum(boolean relative)
	{
		double esqSum  = 0.0;
		
		// simple recursion through non-leaf nodes, adding branches
		if (!isLeaf()) esqSum = left.leafErrorSquaredSum(relative) + right.leafErrorSquaredSum(relative);

		// square the appropriate error vector for leaf objects
		else if (relative) esqSum = relativeStorageError.transpose().times(relativeStorageError).get(0, 0); 
		else if (!relative) esqSum = absoluteStorageError.transpose().times(absoluteStorageError).get(0, 0); 		
		
		return esqSum;
	}
	
	/**
	 * add the computed relative errors to the stored estimates for noisy tracking
	 */
	public void addErrorsToStoredEstimates(int step)
	{
		if (isLeaf()) 
			estimator.locationEstimate[Estimator.NOISYTRACK][step][obj].plusEquals(relativeStorageError);
		
		else if (!isLeaf())
		{
			left.addErrorsToStoredEstimates(step);
			right.addErrorsToStoredEstimates(step);
		}
	}

	
	//---------------------------------------------------------------------------
	//
	//---------------------------------------------------------------------------

	
	/**
	 * 
	 * @param indent
	 */
	public void write(String indent)
	{
		// common for leaves and non-leaves
		String line = indent + "Ani range: " + Util.round(animalRange, 4);
		
		// leaf objects
		if (isLeaf()) 
		{
			line = line + "\tPoint: " + obj + "\tCorr. Esq: " + Util.round(leafErrorSquaredSum(true), obj);
			Util.message(line);
		}
		
		// non-leaf nodes
		else if (!isLeaf())
		{
			line = line + "\tSplit range: " + Util.round(splitRange(), 4);
			Util.message(line);
			
			String nextIndent = "   " + indent;
			left.write(nextIndent);
			right.write(nextIndent);
		}
	}
	
	/**
	 * 
	 */
	public void writeLeafErrors()
	{
		// leaf objects
		if (isLeaf())
		{
			String gap = "---";
			if (parent != null) gap =  Double.valueOf(Util.round(parent.splitRange(), 4)).toString();;
			String line = "Obj: " + obj + " gap: " + gap + "  Rel: ";

			for (int dim = 0; dim < 3; dim++) line = line + "\t" + Util.round(absoluteStorageError.get(dim, 0), 4);
			line = line + "\tAbs: ";
			for (int dim = 0; dim < 3; dim++) line = line + "\t" + Util.round(relativeStorageError.get(dim, 0), 4);

			Util.message(line);
		}
		// non-leaf nodes
		else if (!isLeaf())
		{
			left.writeLeafErrors();
			right.writeLeafErrors();
		}
	}

}
