package mPack;

import java.util.Vector;

import Jama.EigenvalueDecomposition;
import Jama.Matrix;
import display.Reporter;
import utilPack.Util;

public class Estimator {

	private Truth truth;

	private SenseData sense;
	
	private SFM_Model model;
	
	private Reporter reporter;
	
	
	// displayed estimation type
	public int estTypeDrawn = 0;
	
	// types of estimate
	public static int FULLBAYES = 0;
	public static int MEAN = 1;
	public static int TRACK = 2;
	public static int NOISYTRACK = 3;
	
	
	// allowed values of the variable calcMethod	
	public static int START_BAYES = 0;
	public static int RUN_BAYES = 1;
	public static int START_TRACK = 2;
	public static int RUN_TRACK = 3;
	public static int START_NOISY_TRACK = 4;
	public static int RUN_NOISY_TRACK = 5;


	
	// dimension of memory noise: 1 or 3
	public int noiseDimension = 3;
	
	/// true if memory errors for nearby points are correlated
	public boolean relative = false;

	
	// the error in sense data increases by this factor for each extra step in the past 
	public double decay = 1.3;
	
	// the level (x) of memory noise in tracking, from 1 to 7. Noise is 10**(-x) (10**-1.68 is 0.02)
	public double mem_noise  = 1.3;
	
	// number of steps used in the full Bayesian computation
	public double stepsUsed = 5.0;
	
	// threshold for the smallest eigenvalue of the 2nd derivative matrix, to locate an object
	public double threshold = 0.1;

	// maximum number of estimate types
	public static int estTypes = 6;
	
    //--------------------------------------------------------------------------------------
    //     Dynamical model data - animal's estimates, generic for different estimate types 
    //--------------------------------------------------------------------------------------
	
	// true if an object has been placed in any step (do not draw the estimated object location if this is false)
	public boolean[][][] estimated = new boolean[estTypes][SFM_Model.maxSteps][SFM_Model.maxObjects];
	
	// maximum likelihood locations of objects estimated at each step, 3*1 Matrix
	public Matrix[][][] locationEstimate = new Matrix[estTypes][SFM_Model.maxSteps][SFM_Model.maxObjects];
	
	// negative log likelihoods at best fit positions
	public double[][][] negLogLikelihood = new double [estTypes][SFM_Model.maxSteps][SFM_Model.maxObjects];
			
	// second derivative of negative log likelihood
	public Matrix[][][] negLog2Deriv = new Matrix[estTypes][SFM_Model.maxSteps][SFM_Model.maxObjects];

	// 3*3 matrix of error bars around maximum likelihood locations. Each column (second index) = one orthogonal error bar, with length = one standard deviation
	public Matrix[][][] locErrorBars = new Matrix[estTypes][SFM_Model.maxSteps][SFM_Model.maxObjects];

	
    //--------------------------------------------------------------------------------------
    //              Dynamical model data - animal's estimates, old variables 
    //--------------------------------------------------------------------------------------

	
	// Monte Carlo integration points where the negative log likelihood is less than 2.0 more than the minimum
	public Vector<Matrix> likelyPoints = new Vector<Matrix>();
	
	
	//-----------------------------------------------------------------------
	//
	//-----------------------------------------------------------------------

	
	public Estimator(Truth truth, SenseData sense)
	{
		this.truth = truth;
		
		this.sense = sense;
		
	}
	
	public void setModel(SFM_Model model) {this.model = model;}
	
	public void setrReporter (Reporter reporter) {this.reporter = reporter;}
	
	
	   
//--------------------------------------------------------------------------------------
//                    Estimating  object positions
//--------------------------------------------------------------------------------------
	
	/**
	 * @return the ranking of object 0 in negative log likelihood, for different estimation methods
	 * 
	 */
	public void estimatePositions(int step)
	{
		int trace = 0;

		// loop over objects		
		for (int obj = 0; obj < truth.allPoints; obj++)
		{
			// always estimate object position by full Bayes
			estimateObjectPosition(step,obj,trace);
			
			// try to track object position, from previous estimate (tracked, or full Bayes at the start)
			trackObjectPosition(step,obj,trace);
			
			// try to track object position with memory noise, from previous estimate (noisy tracked, or full Bayes at the start)
			noisyTrackObjectPosition(step,obj,trace);
		}
		
		// for noisy tracking, add memory errors in a way that can be done for all points together, for correlated memory errors		
		addMemoryNoise(step);
		
		
		// per step diagnostics
		
		
		// false = do not write mean  estimates (sometimes gives a null)
		// reporter.writeDistances(step,false);
		
		// sometimes freezes
		// reporter.writeErrorBars(step);
		
		// reporter.realityCheck(step);
		
		// reporter.motionCheck(step);
		
	}
	
	/**
	 * 
	 * @param step
	 */
	private void addMemoryNoise(int step)
	{
		if (relative)
		{
			Util.message("\nRelative memory errors\n");
			addRelativeMemoryErrors(step);
		}
		
		else if (!relative)
		{
			for (int obj = 0; obj < truth.allPoints; obj++)
			{
				if (estimated[NOISYTRACK][step][obj])
				{
					Matrix estPosition = locationEstimate[NOISYTRACK][step][obj];

					// add noise to the estimated position, before storing it
					Matrix noisyEstimate = addUnCorrelatedMemoryNoise(estPosition,step,obj);
					locationEstimate[NOISYTRACK][step][obj] = noisyEstimate;
				}
			}
		}
		
	}
	
	/**
	 * 
	 * @param step
	 */
	private void addRelativeMemoryErrors(int step)
	{
		int estimatedPoints = 0;
		for (int obj = 0; obj < truth.allPoints; obj++)
			if (estimated[NOISYTRACK][step][obj]) estimatedPoints++;
		
		
		// if any points have been estimated by noisy tracking
		if (estimatedPoints > 0)
		{
			// current animal position
			Matrix animalPos = Util.makeMatrix(truth.animalPosition[step]);
			
			// binary tree of points to calculate relative errors
			BinaryTree topTree = makeBinaryTree(estimatedPoints,step, animalPos);
			
			// compute scaling factor for relative errors, to make their variances equal to varuance of absolute errors
			double scale = compareRelative(topTree,1.0,step);
			
			// set correlated storage errors top-down, with the correct scale factor
			topTree.findStorageError(scale,step,true);
			
			// write out comparison of correlated and uncorrelated errors
			compareRelative(topTree,scale,step);
			
			//add all correlated errors to noisy tracking position estimates
			topTree.addErrorsToStoredEstimates(step);

		}
	}
	
	/**
	 * 
	 * @param estimatedPoints
	 * @param step
	 * @param animalPos
	 * @return a BinaryTree of all the points that have been estimated by noisy tracking
	 */
	private BinaryTree makeBinaryTree(int estimatedPoints, int step,Matrix animalPos)
	{
		int[] estIndex = new int[estimatedPoints];
		Matrix[] cleanPos = new Matrix[estimatedPoints];

		// collect the estimated positions of all points
		int estPoint = 0;
		for (int obj = 0; obj < truth.allPoints; obj++) if (estimated[NOISYTRACK][step][obj])
		{
			estIndex[estPoint] = obj;
			cleanPos[estPoint] = locationEstimate[NOISYTRACK][step][obj];
			estPoint++;
		}
		
		// arrange the estimates in a binary tree, by proximity			
		// make a flat list of leaf trees
		Vector<BinaryTree> trees = new Vector<BinaryTree>();
		for (int p = 0; p < estimatedPoints; p++)
		{
			BinaryTree leafTree = new BinaryTree(this,animalPos,cleanPos[p],estIndex[p]);
			trees.add(leafTree);				
		}
		
		// make the tree bottom-up in multiple passes
		while (trees.size() > 1)
			trees = nextPass(trees);				
		
		BinaryTree topTree = trees.get(0);
		
		return topTree;
	}
	
	/**
	 * write comparisons of correlated and uncorrelated  memory storage errors
	 * @param topTree
	 * @param scale
	 * @param step
	 */
	private double compareRelative(BinaryTree topTree,double scale, int step)
	{
		// diagnostic write of binary tree
		// topTree.write("");

		// set uncorrelated storage errors  on leaf objects (for comparison of variances)
		topTree.findStorageError(scale,step,false);

		/* compare the two
		Util.message("\nSum of squared errors \nRelative: " + Util.round(topTree.leafErrorSquaredSum(true),4)
		+ "\nAbsolute: "  + Util.round(topTree.leafErrorSquaredSum(false),4) + "\n" );
		topTree.writeLeafErrors();
		*/
		
		// comparison with noise washed out by repetition 1000 times
		double relSum = 0.0;
		double absSum = 0.0;
		for (int i = 0; i < 1000;i++)
		{
			topTree.setZeroErrors();
			// set correlated storage errors top-down
			topTree.findStorageError(scale,step,true);
			// set uncorrelated storage errors  on leaf objects
			topTree.findStorageError(scale,step,false);
			
			relSum = relSum + topTree.leafErrorSquaredSum(true);
			absSum = absSum + topTree.leafErrorSquaredSum(false);				
		}
		
		Util.message("\nGrand sum of squared errors with scale " + Util.round(scale,4)
		+ ": \nRelative: " + Util.round(relSum,4) 
		+ "\nAbsolute: "  + Util.round(absSum,4)
		+ "\nRel/abs " + Util.round((relSum/absSum),4));
		
		double newScale = Math.sqrt(absSum/relSum);
		return newScale;
	}
	
	/**
	 * from one level of the binary tree, compute the next level up,
	 * joining nearest neighbours as much as possible
	 * @param trees
	 * @return
	 */
	private Vector<BinaryTree> nextPass(Vector<BinaryTree>  trees)
	{
		// array of squared distances between centres
		double[][] dist = new double[trees.size()][trees.size()];
		for (int i = 0; i < trees.size();i++)
			for (int j = 0; j < trees.size();j++)
			{
				Matrix ci = trees.get(i).centre;
				Matrix cj = trees.get(j).centre;
				Matrix diff = ci.minus(cj);
				dist[i][j] = Math.sqrt(diff.transpose().times(diff).get(0, 0));
			}
		
		// track which trees from the previous path have been paired up
		boolean[] used = new boolean[trees.size()];
		for (int t = 0; t < trees.size(); t++) used[t] = false;
		int useCount = 0;
		
		// start the list of trees for the next pass
		Vector<BinaryTree> newTrees = new Vector<BinaryTree>();

		// as long a there are two or more trees from the last pass, not yet paired
		while (useCount < trees.size() - 1)
		{
			double minDist = 1000.0;
			int mini = -1;
			int minj = -1;
			
			// find the smallest distance remaining between unpaired trees
			for (int i = 0; i < trees.size();i++) if (!used[i])
				for (int j = i + 1; j < trees.size();j++) if (!used[j])
				{
					if (dist[i][j] < minDist)
					{
						minDist = dist[i][j];
						mini = i;
						minj = j;
					}					
				}
			
			
			BinaryTree newTree = new BinaryTree(this,trees.get(mini),trees.get(minj));
			newTrees.add(newTree);
			useCount = useCount + 2;
			used[mini] = true;
			used[minj] = true;
		}
		
		// pick up one straggler, if there is one
		for (int u = 0; u < used.length; u++) if (!used[u]) newTrees.add(trees.get(u));
		
		return newTrees;
		
	}
	
	
	/**
	 * 
	 * @param estimate
	 * @param step
	 * @param obj
	 * @return
	 */
	private Matrix addUnCorrelatedMemoryNoise(Matrix estimate, int step, int obj)
	{		
		Matrix objectPos = truth.objectLocation[step][obj];
		Matrix noise =  fullMemoryNoise(step, objectPos);

		Matrix estimateWithNoise = estimate.copy().plus(noise);
		return estimateWithNoise;
	}


	/**
	 * 
	 * @param step
	 * @param objectPos object location
	 * 
	 * @return memory noise computed for uncorrelated spatial memory errors,
	 * taking account of the location of the animal at this step
	 */
	public Matrix fullMemoryNoise(int step, Matrix objectPos)
	{		
		Matrix animalPos = Util.makeMatrix(truth.animalPosition[step]);				
		
		// displacement from bee to object
		Matrix diff = objectPos.minus(animalPos);
		double dist = Math.sqrt(diff.transpose().times(diff).get(0, 0));
		
		// noise matrix will be set
		Matrix noise  = new Matrix(3,1);
		
		// compute 10**(-mem_noise)
		double expNoise = Math.exp(-mem_noise*Math.log(10.0));
		
		// 1-D noise along the line of sight
		if (noiseDimension == 1)
		{
			// noise along the line of sight (diff) grows with distance from bee to object, 
			// and has a proportional standard deviation expNoise
			noise  = diff.times(expNoise*Util.randomGaussian());
		}
		
		// 3-D noise in all dimensions, proportional to distance between bee and object
		else if (noiseDimension == 3)
		{
			for (int d = 0; d < 3; d++)
				noise.set(d, 0, dist*expNoise*Util.randomGaussian());
		}
		
		return noise;
		
	}


	
	/**
	 * 
	 * @param step
	 * @param obj
	 * @param trace
	 */
	private void estimateObjectPosition(int step, int obj,int trace)
	{
		int calcMethod = START_BAYES;
		Matrix estPosition = null;
		negLogLikelihood[FULLBAYES][step][obj] = -2.0; // indicates failure, until succeeds
		
		// for comparison, find the negative log likelihood at the true object position
		double[] truePosition = Util.makeVec(truth.objectLocation[step][obj]);
		double trueNegLog = negLogLikelihood(obj,truePosition,null,calcMethod);
		
		// if there is an estimate for the previous step, use that as a starting point
		if (estimated[FULLBAYES][step-1][obj])
		{
			Matrix previous  = locationEstimate[FULLBAYES][step-1][obj]; 
			estPosition = hybridDescent(0,obj,previous,trace,calcMethod);
			
			if (true)
			{
				String result = " failed";
				if (estPosition != null) result = " solved";
				
				double negLog = negLogLikelihood(obj,Util.makeVec(estPosition),null,calcMethod);
				double gap = trueNegLog - negLog;
				
				boolean write = false;
				if (write) Util.message("Object " + obj + result + " at step " + step + 
						", starting from previous estimate." + 
						"\nBest fit neg log likelihood  " + Util.round(negLog, 4)
						+ " and true location value " + Util.round(trueNegLog, 4) + "; gap = " + Util.round(gap, 4));
			}

		}
		
		// if there is no previous estimate, or  if starting from it failed, try random starts
		if (estPosition == null) estPosition = randomStart(obj,step,trace);
		
		// if either of these led to a viable estimate
		if (estPosition != null)
		{
			boolean goodEnough = stableStartPoint(obj,estPosition,true,trace,calcMethod);

			estimated[FULLBAYES][truth.currentStep][obj] =  goodEnough; 
			
			if (goodEnough)
			{
				locationEstimate[FULLBAYES][truth.currentStep][obj] = estPosition;
				Matrix deriv2 = secondDeriv(obj, Util.makeVec(estPosition), calcMethod);
				negLog2Deriv [FULLBAYES][step][obj] = deriv2;
				locErrorBars[FULLBAYES][step][obj] = findErrorBars(deriv2);
				
				negLogLikelihood[FULLBAYES][step][obj] = negLogLikelihood(obj,Util.makeVec(estPosition), null,calcMethod);
				
				// estimated position and variance from the full likelihood distribution
				try
				{
					Matrix[] refined = refineErrors(obj, step, estPosition, calcMethod);
					locationEstimate[MEAN][step][obj] = refined[0];
					locErrorBars[MEAN][step][obj] = refined[1];
				}
				catch (Exception ex) {Util.message(ex.getMessage());}
			}
						
			
		}
		
		else estimated[FULLBAYES][truth.currentStep][obj] = false;
		
	}
	
	/**
	 * 
	 * @param deriv2 matrix of second derivatives of negative log likelihood
	 * @return matrix of standard deviations
	 */
	private Matrix findErrorBars(Matrix deriv2)
	{
		
		Matrix[] eigen = Util.eigenCheck(deriv2, false);
		Matrix eVals = eigen[0];
		Matrix eVecs = eigen[1];
		
		// Util.checkOrtho(eVecs, " second derivative eigenvectors");
		
		Matrix eBars = Matrix.identity(3, 3).times(0.0);

		// loop over eigenvalues of the second derivative matrix
		for (int ev = 0; ev < 3; ev++)
		{
			// one standard deviation along the eigenvector
			double sigma = Math.sqrt(0.5/eVals.get(ev, ev));
			
			// make the error bar vector a multiple of the eigenvector, times sigma
			for (int d = 0; d  < 3; d++) 
				eBars.set(d, ev,sigma*eVecs.get(d, ev));

		}
		
		// Util.checkOrtho(eBars, " error bar matrix");
		
		return eBars;
	}
	

	/**
	 * 
	 * @param step
	 * @param obj
	 * @param trace
	 */
	private void trackObjectPosition(int step, int obj,int trace)
	{
		int calcMethod = -1;
		estimated[TRACK][step][obj] = false; // unless set true later in this method
		negLogLikelihood[TRACK][step][obj] = -2.0; // indicates failure
		Matrix estPosition = null;

		// continue tracking the object if it was successfully tracked in the previous step
		if (estimated[TRACK][step-1][obj])
		{
			calcMethod = RUN_TRACK;
			debugTrace("run track " + obj);

			Matrix previous  = locationEstimate[TRACK][step-1][obj]; 
			estPosition = hybridDescent(0,obj,previous,trace,calcMethod);
		}

		// otherwise, start tracking only if the previous step successfully estimated the object
		else if (estimated[FULLBAYES][step-1][obj])
		{
			calcMethod = START_TRACK;
			debugTrace("start track " + obj);

			// start from the previous Gaussian estmate, not the mean estimate
			Matrix previous  = locationEstimate[FULLBAYES][step-1][obj]; 
			estPosition = hybridDescent(0,obj,previous,trace,calcMethod);
		}
		
		// refine and store results after either START_TRACK or RUN_TRACK
		if (estPosition != null)
		{
			boolean goodEnough = stableStartPoint(obj,estPosition,true,trace,calcMethod);
			
			if (goodEnough)
			{
				// record that tracking has succeeded
				estimated[TRACK][step][obj] = true;
				debugTrace("track errors " + obj);

				locationEstimate[TRACK][step][obj] = estPosition;
				Matrix deriv2 = secondDeriv(obj, Util.makeVec(estPosition), calcMethod);
				negLog2Deriv[TRACK][step][obj] = deriv2;
				locErrorBars[TRACK][step][obj] = findErrorBars(deriv2);
				
				negLogLikelihood[TRACK][step][obj] = negLogLikelihood(obj,Util.makeVec(estPosition), null, calcMethod);
				
			}
			else debugTrace("unstable track " + obj);
			
		}
		else debugTrace("failed track " + obj);
		
	}

	/**
	 * 
	 * @param step
	 * @param obj
	 * @param trace
	 */
	private void noisyTrackObjectPosition(int step, int obj,int trace)
	{
		int calcMethod = -1;
		estimated[NOISYTRACK] [step][obj] = false; // unless set true later in this method
		negLogLikelihood[NOISYTRACK][step][obj] = -2.0; // indicates failure
		Matrix estPosition = null;

		// continue tracking the object if it was successfully tracked in the previous step
		if (estimated[NOISYTRACK][step-1][obj])
		{
			calcMethod = RUN_NOISY_TRACK;
			debugTrace("run noisy track " + obj);

			Matrix previous  = locationEstimate[NOISYTRACK][step-1][obj]; 
			estPosition = hybridDescent(0,obj,previous,trace,calcMethod);
		}

		// otherwise, start tracking only if the previous step successfully estimated the object
		else if (estimated[FULLBAYES][step-1][obj])
		{
			calcMethod = START_NOISY_TRACK;
			debugTrace("start noisy track " + obj);

			// start from the previous Gaussian estimate, not the mean estimate
			Matrix previous  = locationEstimate[FULLBAYES][step-1][obj]; 
			estPosition = hybridDescent(0,obj,previous,trace,calcMethod);
		}
		
		// refine and store results after either START_NOISY_TRACK or RUN_NOISY_TRACK
		if (estPosition != null)
		{
			boolean goodEnough = stableStartPoint(obj,estPosition,true,trace,calcMethod);
			
			if (goodEnough)
			{
				// record that tracking has succeeded
				estimated[NOISYTRACK][step][obj] = true;
				debugTrace("noisy track errors " + obj);
				
				// DO NOT add noise to the estimated position, before storing it; that is now doen for all objects in estPositions
				// Matrix noisyEstimate = addMemoryNoise(estPosition,step,obj);
				// locationEstimate[NOISYTRACK][step][obj] = noisyEstimate;

				// store position estimate with no memory noise - noise is added later
				locationEstimate[NOISYTRACK][step][obj] = estPosition;
				
				// calculate second derivatives at the noisy position? Or the true minimum? Now it has to be the latter
				Matrix deriv2 = secondDeriv(obj, Util.makeVec(estPosition), calcMethod);
				negLog2Deriv[NOISYTRACK][step][obj] = deriv2;
				locErrorBars[NOISYTRACK][step][obj] = findErrorBars(deriv2);

				negLogLikelihood[NOISYTRACK][step][obj] = negLogLikelihood(obj,Util.makeVec(estPosition), null, calcMethod);
				
			}
			else debugTrace("unstable noisy track " + obj);
			
		}
		else debugTrace("failed noisy track " + obj);
		
	}
	
	
	/**
	 * Take N random start points in a cube of size cubeSize about the real object position
	 * @param obj
	 * @param step
	 * @return
	 */
	Matrix randomStart(int obj, int step, int trace)
	{
		int calcMethod = START_BAYES;

		// maximum number of random start points
		int randoms = 10;
		// side of  cube for random starting points
		double cubeSize = 0.5;
		
		Vector<Double> minima = new Vector<Double>();
		Vector<double[]> positions = new Vector<double[]>();
		
		
		Matrix bestRef = null;
		double bestLik = 100000;
		int bestSol = -1;
				
		if (trace == 5) Util.message("\n\n**** " + randoms + " random starting points for object " + obj + " at step " + step);
		
		// random different start points
		for (int r = 0; r < randoms; r++)
		{
			// distance from random start point to true position of object
			double realGap = 0.0;
			double[] rand  = new double[3];			
			for (int dim = 0; dim < 3; dim++)
			{
				double jit = cubeSize*(Math.random() - 0.5);
				rand[dim] = truth.objectLocation[step][obj].get(dim, 0) + jit;
				realGap = realGap + jit*jit;
			}
			realGap = Math.sqrt(realGap);
			if (trace == 5) Util.message("\nStart point " + r + " for object " + obj + " with real position gap " + realGap);
			
			// iterative refinement from this start point
			Matrix lastRef = hybridDescent(r,obj,Util.makeMatrix(rand),trace,calcMethod);
			
			if (lastRef != null)
			{
				boolean stable = stableStartPoint(obj,lastRef,false,trace,calcMethod);
				
				if (stable)
				{
					double logLik = negLogLikelihood(obj,Util.makeVec(lastRef),null, calcMethod);
					minima.add(Double.valueOf(logLik));
					positions.add(Util.makeVec(lastRef));

					if (trace == 5) Util.message("Solution for start " + r + ": " + logLik + "\n");
					
					if (logLik < bestLik)
					{
						bestLik = logLik;
						bestRef = lastRef;
						bestSol = r + 1;
					}
				}
				
			}
			else if (trace == 5) Util.message("Failed random start point " + r + "\n");
		}
		
		if ((bestSol > -1) & (trace == 5)) Util.message("Best solutions at " + bestSol + " random starts, with negative log likelihood " + bestLik );
		
		// histogram log likelihoods, and find variance of estimated positions
		analyse(randoms,obj,minima,positions);
		
		return bestRef;
	}
	
	/**
	 * analysis solutions to reveal spurious local minima
	 * @param minima
	 */
	private void analyse(int randoms, int obj, Vector<Double> minima,Vector<double[]> positions)
	{
		double minLoc = 1000.0;
		double maxLoc = -1000.0;
		double[] meanPosition = new double[3];
		for (int dim = 0; dim < 3; dim++) meanPosition[dim] = 0.0;
		
		
		for (int i = 0; i < minima.size(); i++)
		{
			// find minimum and maximum of log likelihood
			Double dB = minima.get(i);
			double d = dB.doubleValue();
			if (d < minLoc) minLoc = d;
			if (d > maxLoc) maxLoc = d;
			
			// build up the mean position of all solutions
			double[] pos = positions.get(i);
			for (int dim = 0; dim < 3; dim++) 
				meanPosition[dim] = meanPosition[dim] + pos[dim] ;
		}
		double range = maxLoc - minLoc;
		for (int dim = 0; dim < 3; dim++) meanPosition[dim] = meanPosition[dim]/minima.size();
		
		// compute the variance of positions about the mean position
		double variance = 0.0;
		for (int i = 0; i < minima.size(); i++)
		{
			for (int dim = 0; dim < 3; dim++)
			{
				double dev = meanPosition[dim] - positions.get(i)[dim];
				variance = variance + dev*dev;
			}
		}
		
		if (minima.size() > 0)
		{
			variance = variance/minima.size();
			double std = Math.sqrt(variance);

			// Util.message(minima.size() +  "/" + randoms + " stable solutions for object " + obj + " with likelihood range " + range + " from minimum " + minLoc + " and position spread " + std);

			// possible histogram solution log likelihood values
			boolean showHist = false;
			if (showHist & (range > 0.1))
			{
				int bins = 30;
				int[] hist = new int[bins];
				for (int b = 0; b < bins; b++) hist[b] = 0;
				
				for (Double dB:minima)
				{
					double d = dB.doubleValue();
					int bin = (int)(bins*(d-minLoc)/range);
					if (bin == bins) bin--;
					if (bin < 0) bin++;
					hist[bin]++;
				}
				
				String line = "";
				for (int b = 0; b < bins; b++) line = line + hist[b] + ",";
				Util.message(line);			
			}

			
		}
		else Util.message("No stable solutions for object " + obj);
		
	}
	
	
	
	Matrix hybridDescent(int randStart, int obj,Matrix startPosition,int trace, int calcMethod)
	{		
		// iterate until the slope is less than this
		double criterion = 0.0001;
		
		// maximum number of trials before failing
		int maxTrials = 100; 
		int trials = 0;
		
		boolean converged = false;
		
		// to be refined 
		Matrix refined = startPosition.copy();
		
		// iterative loop
		String line = "";
		String type = "";
		while ((refined != null) && (!converged) && (trials < maxTrials))
		{
			trials++;
			
			double slope = slope(obj,refined,calcMethod);
			line = line + "\t" + Util.round(slope,4);
			
			converged = (slope < criterion);
			
			if (!converged)
			{
				Matrix vFirst = firstDeriv(obj, Util.makeVec(refined),calcMethod);			
				Matrix second = secondDeriv(obj, Util.makeVec(refined),calcMethod);

				double smallest = largestEigenvalue(obj, refined,calcMethod);				
				if (useMatrix(slope,smallest, trials))
				{
					type = type + "M";
					refined = matrixStep(trials, obj, refined, vFirst, second,trace,calcMethod);
				}
				
				else
				{
					type = type + "S";
					refined = steepStep(trials, obj, refined, vFirst, second,trace,calcMethod);
				}
			}
		}
		
		if (converged & (trace == 1)) Util.message("Start " + randStart + "; " + trials + " trials of type " + type + line);
		
		if (!converged) refined = null;
		return refined;
	}
	

	
	/**
	 * One step in refining an estimate of position
	 * @paraM 	r refinement step number
	 * @param obj index of the object
	 * @param estimate the input estimate of its position
	 * @param first first derivatives of negative log likelihood wrt position
	 * @param secondDeriv second derivatives
	 * @return
	 */

	Matrix steepStep(int r, int obj, Matrix estimate, Matrix first, Matrix secondDeriv, int trace, int calcMethod)
	{
		// to be returned
		Matrix newEstimate  = new Matrix(3,1);
		double slope  = slope(obj,estimate,calcMethod);
		
		// unit vector in the direction of steepest descent
		Matrix unit = first.times(1.0/slope);
		
		// Matrix product u transpose * M * u; thus is the second derivative in direction of movement
		double utMu = unit.transpose().times(secondDeriv).times(unit).get(0, 0);
		
		// recommended distance to travel - only for trace
		double dist = slope/utMu;
		
		//scale down the distance of travel by factors of 2 until there is improvememt
		boolean improved = false;
		double scale = 1.0;
		int maxTries = 10;
		int tries = 0;
		while ((tries < maxTries) && (!improved))
		{
			newEstimate  = estimate.minus(first.times(scale/utMu));
			
			// component of the new first derivative along the old unit slope vector - should always be able to converge it.
			double newSlopeComponent  = firstDeriv(obj,Util.makeVec(newEstimate),calcMethod).transpose().times(unit).get(0, 0);		
			improved = (newSlopeComponent < slope);
			
			if (!improved)
			{
				scale = scale/2.0;
				tries++;			
			}			
		}
		
		// all attempts failed to get a smaller slope
		if (tries > maxTries -1)  
		{
			Util.message("Failed steepest step");
			newEstimate = null;
		}
		
		
		return newEstimate;
	}
	
	
	/**
	 * One step in refining an estimate of position
	 * @paraM 	r refinement step number
	 * @param obj index of the object
	 * @param estimate the input estimate of its position
	 * @param first first derivatives of negative log likelihood wrt position
	 * @param secondDeriv second derivatives
	 * @return
	 */
	Matrix matrixStep(int r, int obj, Matrix estimate, Matrix first, Matrix secondDeriv, int trace, int calcMethod)
	{
		// do not attempt a matrix iteration if the eigenvectors are not all sufficiently positive
		if (largestEigenvalue(obj,estimate,calcMethod) < 5) return null;

		// invert the matrix of second derivatives
		Matrix inverse = secondDeriv.inverse();
		
		// displacement to first new estimate (points away from the best estimate point)
		Matrix disp = inverse.times(first);
		Matrix newEstimate = estimate.minus(disp);
		
		return newEstimate;
					
	}
	//---------------------------------------------------------------------------------------------------------------
	//                            Negative log likelihoods and derivatives
	//---------------------------------------------------------------------------------------------------------------
	

	/* allowed values of the variable calcMethod	
	public static int START_BAYES = 0;
	public static int RUN_BAYES = 1;
	public static int START_TRACK = 2;
	public static int RUN_TRACK = 3;
	public static int START_NOISY_TRACK = 4;
	public static int RUN_NOISY_TRACK = 5;
	*/

	
	/**
	 * @param obj
	 * @param estimate any estimated position of the object
	 * @param displacement: if not null, add to the previous estimated position when tracking
	 * @return negative log likelihood, summed over all steps 
	 * Contributions of old steps are decayed.
	 */

	public double negLogLikelihood(int obj, double[] estimate, Matrix displacement, int calcMethod)
	{
		// special object for controlled testing
		if (obj == -1) return testObject(estimate);
		
		int previousStep = truth.currentStep - 1;
		
		Matrix nowEst = Util.makeMatrix(estimate);
		
		double decayFactor = 1.0;
		
		// likelihood from current step sense data - always used, for all calculations
		double negLog = senseNegLogLikelihood(truth.currentStep,obj, estimate);
		
		// full Bayesian calculation - add  sense data from all previous steps
		if ((calcMethod == START_BAYES)||(calcMethod == RUN_BAYES))
		{
			// negative log likelihood from previous steps
			double prevNegLog = 0.0;
			
			//   use sense data from  some maximum number of previous steps, back to the zeroth step
			int useStep = Math.min(truth.currentStep,(int)stepsUsed);

			for (int st = 0; st < useStep; st++) // number of steps back into thepast
			{
				int step = truth.currentStep - 1 - st ; // step goes from (truth.currentStep -1) down to zero

				// decay is > 1.0, so past steps count for less; decayFactor keeps shrinking as you go back
				decayFactor = decayFactor/decay;
				
				prevNegLog = prevNegLog + decayFactor*senseNegLogLikelihood(step,obj, estimate);				
			}
			
			negLog = negLog + prevNegLog;
		}
		
		// Start tracking using current step sense data and the previous Gaussian estimate
		else if (calcMethod == START_TRACK)
		{
			decayFactor = 1.0/decay;
			
			double previousNLL = negLogLikelihood[FULLBAYES][previousStep][obj];
			Matrix previousEst = locationEstimate[FULLBAYES][previousStep][obj];
			Matrix previous2Deriv = negLog2Deriv [FULLBAYES][previousStep][obj];
			

			negLog = previousNLL + negLog + decayFactor*estTrackNegLog(previousEst, previous2Deriv,nowEst);
		}
		
		
		// continue tracking using current step sense data and the previous tracked estimate
		else if (calcMethod == RUN_TRACK)
		{
			decayFactor = 1.0/decay;
			
			double previousNLL = negLogLikelihood[TRACK][previousStep][obj];
			Matrix previousEst = locationEstimate[TRACK][previousStep][obj].copy();
			// for motion tracking, add a displacement to the previous estimate (but do not alter the stored estimate - only a copy!)
			if (displacement != null) previousEst.plusEquals(displacement);
			Matrix previous2Deriv = negLog2Deriv[TRACK][previousStep][obj];

			negLog = previousNLL + negLog + decayFactor*estTrackNegLog(previousEst, previous2Deriv, nowEst);
		}

		
		// Start tracking with memory noise, using current step sense data and the previous Gaussian estimate
		// (the first step of noisy tracking has no difference from tracking with no noise)
		else if (calcMethod == START_NOISY_TRACK)
		{
			decayFactor = 1.0/decay;
			
			double previousNLL = negLogLikelihood[FULLBAYES][previousStep][obj];
			Matrix previousEst = locationEstimate[FULLBAYES][previousStep][obj];
			Matrix previous2Deriv = negLog2Deriv [FULLBAYES][previousStep][obj];
			
			double senseNegLog = negLog;

			negLog = previousNLL + senseNegLog + decayFactor*estTrackNegLog(previousEst, previous2Deriv,nowEst);
		}
		
		
		// continue tracking with memory noise, using current step sense data and the previous noisy tracked estimate
		else if (calcMethod == RUN_NOISY_TRACK)
		{
			decayFactor = 1.0/decay;
			
			double previousNLL = negLogLikelihood[NOISYTRACK][previousStep][obj];
			Matrix previousEst = locationEstimate[NOISYTRACK][previousStep][obj].copy();
			// for motion tracking, add a displacement to the previous estimate (but do not alter the stored estimate - only a copy!)
			if (displacement != null) previousEst.plusEquals(displacement);
			Matrix previous2Deriv = negLog2Deriv[NOISYTRACK][previousStep][obj];

			negLog = previousNLL + negLog + decayFactor*estTrackNegLog(previousEst, previous2Deriv, nowEst);
		}

		
		return negLog;
	}
	

	
	/**
	 * 
	 * @param previousEst previous central estimate of position
	 * @param deriv2 previous matrix of second derivatives of negative log likelihood
	 * @param estMatrix trial position being tested
	 * @return
	 */
	private double estTrackNegLog( Matrix previousEst, Matrix  deriv2, Matrix estMatrix)
	{
		Matrix diff = estMatrix.minus(previousEst);
			
		double negLog = 0.5*diff.transpose().times(deriv2.times(diff)).get(0, 0);
		
		return negLog;
	}

	
	/**
	 * 
	 * @param step the step 0... currentStep
	 * @param obj the object observed by the animal, with errors
	 * @param estimate any estimated position
	 * @return negative log likelihood contribution for the object in the step
	 */
	private double senseNegLogLikelihood(int step, int obj, double[] estimate)
	{
		double negLog = 0.0;
		
		// displacement from the animal position to the estimate, and distance
		double[] disp = new double[3];
		double dist = 0.0;
		for (int dim = 0; dim < 3; dim++)
		{
			disp[dim] = estimate[dim] - truth.animalPosition[step][dim];
			dist = dist + disp[dim]*disp[dim];			
		}
		dist = Math.sqrt(dist);
		
		if (model.species == SFM_Model.BEE)
		{
			// unit vector in the direction of the object, with errors
			double[] unitV = sense.beeVisionData[step][obj];
			
			// cosine of the angle between the displacement and the unit vision vector
			double cosAng = 0.0;
			for (int dim = 0; dim < 3; dim++)
				cosAng = cosAng + disp[dim]*unitV[dim];
			cosAng = cosAng/dist;
			
			// sin  of the angle (sqrt may blow up from rounding errors)
			double sin2 = 1.0 - cosAng*cosAng;
			double sinAng = 0.0;
			if (sin2 > 0.0) sinAng= Math.sqrt(sin2);
			
			// convert from radians (sin) to  degrees, assuming sinAng is small
			sinAng = sinAng*90.0/Math.PI;
			
			// (angle/bee error) squared. bee error is one standard deviation
			negLog = sinAng*sinAng/(2.0*sense.err_bee*sense.err_bee);
		}
		
		
		if (model.species == SFM_Model.BAT)
		{
			// sonar data for this step, this object
			double[] echoData = sense.batSonarData[step][obj];
			
			// radius of the echo sphere for this object
			double radius = echoData[3];

			// how far the estimated position is off the surface of the echo sphere, in or out
			double farOut = radius - dist;
			
			// range error contribution - (farOut/range error) squared - range_err is one standard deviation
			negLog = farOut*farOut/(2.0* sense.range_err*sense.range_err);
			
			// unit vector along the bat's velocity
			double[] unit = new double[3];
			for (int dim = 0; dim <3; dim++) unit[dim] = echoData[dim];
					
			// cosine of the angle between the unit direction vector, and the displacement from the bat to the estimated position
			double cosAlpha= 0.0;
			for (int dim = 0; dim <3; dim++) cosAlpha = cosAlpha + unit[dim]*disp[dim]/dist;
			
			// doppler error contribution.
			double cosTheta = echoData[4];
			double dCos = cosAlpha - cosTheta;
			// doppler_err is one standard deviation
			negLog = negLog + dCos*dCos/(2.0* sense.doppler_err*sense.doppler_err);
		}

		return negLog;
	}
	
	/**
	 * @param obj
	 * @param estimate
	 * @return second derivatives of negative log likelihood wrt estimate positions
	 */
	public Matrix secondDeriv(int obj, double[] estimate, int calcMethod)
	{
		double tiny = 0.00001;
		Matrix second = new Matrix(3,3);

		// make a small change in each dimension and measure the difference		
		for (int dim = 0; dim < 3; dim++)
		{
			double[] plusPoint = new double[3];
			double[] minusPoint = new double[3];
			for (int d = 0; d < 3; d++)
			{
				plusPoint[d] = estimate[d];
				minusPoint[d] = estimate[d];
			}
			plusPoint[dim] = plusPoint[dim] + tiny/2.0;
			minusPoint[dim] = minusPoint[dim] - tiny/2.0;
			
			Matrix slopes = (firstDeriv(obj, plusPoint,calcMethod).minus(firstDeriv(obj, minusPoint,calcMethod))).times(1.0/tiny);
			for (int d = 0; d < 3; d++)
				second.set(dim,d,slopes.get(d,0));
		}

		
		// make sure the second derivative matrix is precisely symmetric, and check it is approximately symmetric
		Matrix secondSym = new Matrix(3,3);
		// Util.message("\nChecking symmetry for object " + obj);
		for (int d = 0; d < 3; d++)
		{
			String line = "";
			for (int dp = 0; dp < 3; dp++)
			{
				secondSym.set(d, dp, 0.5*(second.get(d,dp) + second.get(dp, d)));
				// check asymmetry is near zero
				double asym = 0.5*(second.get(d,dp) - second.get(dp, d));
				line = line + Double.valueOf(asym).doubleValue() + "\t";				
			}
			// Util.message(line);
		}
		
		return secondSym;
	}
	
	/**
	 * @param obj
	 * @param estimate
	 * @return 3*1 Matrix (column vector) 
	 * the first derivative of the negative log likelihood at the estimated point
	 */
	public Matrix firstDeriv(int obj, double[] estimate, int calcMethod)
	{
		double tiny = 0.00001;
		Matrix first = new Matrix(3,1);
		
		// make a small change in each dimension and measure the difference
		for (int dim = 0; dim < 3; dim++)
		{
			double[] plusPoint = new double[3];
			double[] minusPoint = new double[3];
			for (int d = 0; d < 3; d++)
			{
				plusPoint[d] = estimate[d];
				minusPoint[d] = estimate[d];
			}
			plusPoint[dim] = plusPoint[dim] + tiny/2.0;
			minusPoint[dim] = minusPoint[dim] - tiny/2.0;
			
			double slope = (negLogLikelihood(obj, plusPoint,null, calcMethod) - negLogLikelihood(obj, minusPoint,null, calcMethod))/tiny;
			first.set(dim, 0, slope);
		}
		
		return first;
	}
		
	
	//-------------------------------------------------------------------------------------------
	//         Computation of mean positions and true errors by looking at the negative log likelihood
	//-------------------------------------------------------------------------------------------
	
	/**
	 * 
	 * @param obj
	 * @param gaussianEstimate
	 * 
	 * @return 2 matrices: [mean position estimate, error bars]
	 */
	private Matrix[] refineErrors(int obj, int step, Matrix gaussianEstimate, int calcMethod) throws Exception
	{
		// second derivatives in the Gaussian approximation
		Matrix second = secondDeriv(obj, Util.makeVec(gaussianEstimate), calcMethod);
		
		// Monte Carlo integration of actual likelihood over 5 standard deviations of the Gaussian approximation (also makes 'likely points' cloud)
		// use a small number of points beause this is no longer very interesting
		int montePoints = 500;
		Matrix[] result = findTrueError(obj,gaussianEstimate,second, montePoints,0.5, false, calcMethod);
		
		// moments of the actual likelihood distribution
		double zeroMoment = result[0].get(0,0);
		Matrix firstMoment = result[1];
		Matrix secondMoment = result[2];
		
		/// normalise first and second moments to 1.0
		Matrix firstNorm = firstMoment.times(1.0/zeroMoment);
		Matrix secondNorm = secondMoment.times(1.0/zeroMoment);
		
		// mean position of the object in the likelihood distribution
		Matrix meanEstimate = gaussianEstimate.plus(firstNorm);
		
		// second derivatives around the mean position
		Matrix firstOuter = firstNorm.times(firstNorm.transpose());
		Matrix variance = secondNorm.minus(firstOuter);
		
		// Eigenvalues and eigenvectors of the variance matrix; true to write out a check
		Matrix[] eigen = Util.eigenCheck(variance, false);
		Matrix eVals = eigen[0];	
		Matrix eVecs = eigen[1];
		
		// loop over eigenvalues
		Matrix eBar = new Matrix(3,3);
		Matrix sDevs = new Matrix(3,1); // standard deviations, for diagnosie
		Matrix longest = new Matrix(3,1); // longest error bar for diagnosis
		double maxDev = 0.0;
		for (int ev = 0; ev < 3; ev++)
		{
			// one standard deviation along this eigenvector
			double sDev = Math.sqrt(eVals.get(ev, ev));
			sDevs.set(ev, 0, sDev);
			
			// change eigenvectors from norm 1 to norm one standard deviation
			for (int dim = 0; dim < 3; dim++) // loop over x,y,z
				eBar.set(dim, ev, sDev*eVecs.get(dim, ev));
			
			// record the longest error bar
			if (sDev > maxDev)
			{
				maxDev = sDev;
				for (int dim = 0; dim < 3; dim++)
					longest.set(dim, 0, eBar.get(dim, ev));
			}
		}
		
		
		// diagnostics
		/*
		Matrix animal = makeMatrix(animalPosition[step]);
		Matrix object = makeMatrix(truth.objectLocation[obj]);

		write2(object, animal,"\nObject","Animal");
		write2(gaussianEstimate,object,"\nGauss est.","Object");		
		write2(meanEstimate,object,"\nMean est.","Object");	
		
		write1(sDevs,"\nStand. devs");
		write1(longest,"Long err bar");
		*/

		Matrix[] res = new Matrix[2];
		res[0] = meanEstimate;
		res[1] = eBar;
		return res;
		
	}
	
	public static void write2(Matrix vec1,Matrix vec2, String title1, String title2)
	{
		write1(vec1,title1);
		write1(vec2,title2);
		write1(vec1.minus(vec2),"difference");
	}
	
	/**
	 * 
	 * @param vec
	 * @param title
	 */
	public static void write1(Matrix vec, String title)
	{
		String line = title;
		for (int d = 0; d < 3; d++)
			line = line + "\t" + Util.round(vec.get(d, 0), 6);
		Util.message(line);
	}

	
	
	/**
	 * controlled test of method findTrueError
	 * with the special object -1 having controllable negLogLikelihood
	 */
	public void doControlledTest() throws Exception
	{
		// special object for controlled testing
		int obj = -1;
		
		// Gaussian 'ideal' central best estimate (x,y,z) = (1,0,0)
		Matrix best = new Matrix(3,1);
		best.set(0, 0, 1.0);
		best.set(1, 0, 0.0);
		best.set(2, 0, 0.0);
		
		// Gaussian 'ideal' second derivative matrix
		Matrix deriv2 = Matrix.identity(3, 3).times(10.0);
				
		// single test of controlled object, with detailed write out
		Matrix[] result = findTrueError(obj,best,deriv2,80000,1.0, true, 0);
		
		// show normalised variance matrix
		analyseResult(result);
		
		// repeat with vmin = 0.2; should give the same result
		result = findTrueError(obj,best,deriv2,80000,0.4, true, 0);
		analyseResult(result);

		// test a range of vMin values, from 1.0 downwards
		for (int vb = 0; vb < 0; vb++)
		{
			double vMin = 1.0 - 0.1*(double)vb;
			int montePoints = 80000;
			Util.message("\nTest with vMin = " + (10 - vb) + "/10; and " + montePoints + " integration points.");
			
			String zeroLine = "";
			String oneLine = "";
			double[] track = {0.0,0.0,0.0,0.0};
			// for each vMin, replicate 12 times to compute the variances of zeroth and first moments
			for (int testNo = 0; testNo < 12; testNo++)
			{
				Matrix[] res = findTrueError(obj,best,deriv2,16000,vMin, false, 0);

				// zeroth moment divisor
				double zeroM = res[0].get(0, 0);
				zeroLine = zeroLine + Util.round(zeroM, 4) + "\t";
				// first moment x position
				double oneM = res[1].get(0, 0);
				oneLine = oneLine + Util.round(oneM, 4) + "\t";
				
				track[0] = track[0] + zeroM; // mean of zeroth moment
				track[1] = track[1] + zeroM*zeroM; //variance of zeroth moment
				track[2] = track[2] + oneM; // mean of first moment
				track[3] = track[3] + oneM*oneM; //variance of first moment
			}
			double mean0 = track[0]/12.0;
			double var0 = track[1]/12.0 - mean0*mean0;
			double mean1 = track[2]/12.0;
			double var1 = track[3]/12.0 - mean1*mean1;

			Util.message(zeroLine + "; variance " + var0);
			Util.message(oneLine + "; variance " + var1);
			
		}
		
		
	}
	
	/**
	 * estimate of negative log likelihood, for
	 * special object -1 for controlled testing
	 * @param estimate random test position in Monte Carlo integration
	 * @return
	 */
	private double testObject(double[] estimate)
	{
		// centre of the test object (x,y,z)
		double[] realCentre = {1.2,0.0,0.0};
		
		// displacement from the centre point (x,y,z)
		Matrix disp = Util.makeMatrix(estimate).minus(Util.makeMatrix(realCentre));
		
		// zero first derivative at the centre point realCentre

		// second derivative of  negative log likelihood at the test centre point
		Matrix m2 = Matrix.identity(3, 3).times(10.0);
		// decrease precision in y
		m2.set(1, 1, 5.0);
		// increase precision in x
		m2.set(0, 0, 40.0);
		
		Double negLog = disp.transpose().times(m2.times(disp)).get(0, 0);
		
		return negLog;
	}
	
	/**
	 * analyse results of controlled test
	 */
	private void analyseResult(Matrix[] result)
	{
		String[] type = {"Actual","Gaussian"};
		// loop over Actual/Gaussian check
		for (int rc = 0; rc < 2; rc++)
		{
			int realCheck = 3*rc;
			double zeroMoment = result[realCheck].get(0,0);
			Matrix firstMoment = result[realCheck + 1];
			Matrix secondMoment = result[realCheck + 2];
			
			writeMoments(type[rc], zeroMoment,firstMoment,secondMoment);
			
			/// normalise to 1.0
			Matrix firstNorm = firstMoment.times(1.0/zeroMoment);
			Matrix secondNorm = secondMoment.times(1.0/zeroMoment);
			
			// second derivatives around the mean
			Matrix firstOuter = firstNorm.times(firstNorm.transpose());
			Matrix secondCentred = secondNorm.minus(firstOuter);
			
			writeMoments(type[rc] + " Centred", 1.0,firstNorm,secondCentred);
			
		}
	}

	/**
	 * simple write out of moments
	 * @param type
	 * @param zeroMoment
	 * @param firstMoment
	 * @param secondMoment
	 */
	private void writeMoments(String type, double zeroMoment,Matrix firstMoment,Matrix secondMoment)
	{
		Util.message("\n" + type + " zeroth moment: " + Util.round(zeroMoment, 6));
		
		for (int dim = 0; dim < 3; dim ++)
		{
			String line = dim + "\t" + Util.round(firstMoment.get(dim,0), 6) + "\t";
			for (int d = 0; d < 3; d++) 
				line = line + "\t" + Util.round(secondMoment.get(dim, d), 6);
			Util.message(line);
		}
	}

	
	/**
	 * 
	 * @param standardDevs
	 * @param pBin
	 * @param pureGauss
	 * @param actual
	 */
	private void writeHistograms(int montePoints, double standardDevs, int[] pBin, double[] pureGauss, double[]actual)
	{
		String header = "\nS.D.\tg\tPoints\t\tTheoryG\tSumG\tEachG\texp(-g)\t\tSumAct\tEachAct";
		Util.message(header);

		int bins = pBin.length;
		for (int b = 0; b < bins; b++) 
		{
			double x = standardDevs*((double)b)/((double)bins);
			double g = 0.5*x*x;
			double expg = Math.exp(-g);
			double theoryGaussBin = 4.0*Math.PI*x*x*Math.exp(-x*x/2.0);

			Util.message
			(Util.round(x, 2) + "\t"  									// standard deviation
			+ Util.round(g, 4) + "\t"                  					// g
			+ pBin[b] + "\t\t"  										// integration points in bin
			+ Util.round(theoryGaussBin, 4)+ "\t"  						// theory Gaussian
			+ Util.round(pureGauss[b], 4) + "\t"  						// summed Gaussian
			+ Util.round(pureGauss[b]*montePoints/pBin[b], 4) + "\t" 	// pure Gaussian per point
			+ Util.round(expg, 4) + "\t\t"                  			// exp(-g)
			+ Util.round(actual[b], 4) + "\t" // summed actual			// summed actual
			+ Util.round(actual[b]*montePoints/pBin[b], 4) + "\t");		// actual per point
		}
	}
	
	
	/**
	 * the zeroth moment integral of a pure gaussian likelihood 
	 * @param deriv2
	 * @return
	 */
	private double gaussianZeroMoment(Matrix deriv2, boolean write)
	{ 
		double volume = -1.0; //absurd answer
		
		// product of three  eigenvalues should be positive
		EigenvalueDecomposition eComp = deriv2.eig();
		Matrix eVals = eComp.getD();		
		double eValProd = eVals.get(0,0)*eVals.get(1,1)*eVals.get(2,2);
		
		String line = "\nEigenvalues: \t";
		for (int dim = 0; dim < 3; dim++)
			line = line + Util.round(eVals.get(dim, dim),2) + "\t";
		if (write) Util.message(line + "; product: " + Util.round(eValProd, 2));
		
		line = "One standard deviation gaussian error bars: ";
		double rectVol = 8.0; // 2 s.d. each dimension
		for (int dim = 0; dim < 3; dim++)
		{
			double eBar = Math.sqrt(0.5/eVals.get(dim, dim));
			rectVol = rectVol*eBar;
			line  = line + Util.round(eBar, 6) + "\t";		
		}
		if (write) Util.message(line + "; rectangular volume: " + rectVol);

		if (eValProd > 0.0)
		{
			double piCubed = Math.PI*Math.PI*Math.PI;
			volume = Math.sqrt(piCubed/eValProd);
		}
		
		if (write) Util.message("Theoretical Gaussian volume: " + volume);
		
		return volume;
		
	}

	// ---------------------------------------------------------------------------------------------------
	//                              Little-used methods	
	// ---------------------------------------------------------------------------------------------------


	/**
	 * 
	 * @param objx object index (-1 for controlled test)
	 * @param bestEstimate maximum likelihood position estimate
	 * @param deriv2 second derivative of negative log likelihood at the maximum
	 * @param montePoints number of Monte Carlo integration points
	 * @param vMin value of v = sqrt(g/gmax), above which there are fewer points with higher weight
	 * 
	 * @return 
	 */
	private Matrix[] findTrueError(int objx, 
			Matrix bestEstimate, 
			Matrix deriv2, 
			int montePoints, 
			double vMin, 
			boolean write, 
			int  calcMethod)  throws Exception
	{
		// parameters to be used

		//number of standard deviations to integrate out to
		double standardDevs = 5.0;
		// bins for points, uniformly spaced in standard deviations
		int  bins = 10;
		double dBins = (double)bins;
				
		// expected zeroth moment integral for Gaussian likelihood
		double trueVolume = gaussianZeroMoment(deriv2, write);
		
		// re-initialise the set of points with negative log likelihood within 3.0 of the minimum
		likelyPoints = new Vector<Matrix>();

		
		// negative log likelihood at the maximum likelihood point
		double centreNegLog = negLogLikelihood(objx, Util.makeVec(bestEstimate), null,calcMethod);
		
		// make the tilted 3-D rectangle within which points are to be generated
		Matrix[] eigen = Util.eigenCheck(deriv2, false);
		Matrix[] side = new Matrix[3]; // 3 sides of the rectangle (3-vectors)
		Matrix corner = Matrix.identity(3, 1).times(0.0); // start corner, relative to maximum likelihood point,  for generating points (3-vector)
		
		// loop over eigenvalues of the second derivative matrix
		for (int ev = 0; ev < 3; ev++)
		{
			// one standard deviation along this eigenvector
			double sDev = 1.0/Math.sqrt(2.0*eigen[0].get(ev, ev));

			// vector defining one side of the rectangle (by scaling the unit eigenvector)
			side[ev] = new Matrix(3,1);
			for (int dim = 0; dim < 3; dim++)
				side[ev].set(dim, 0, standardDevs*sDev*eigen[1].get(dim,ev));
			
			// displace the start corner by minus this vector (5 standard deviations)
			corner = corner.minus(side[ev]);
		}
		
		// all integration points
		Vector<Matrix> monteCarloPoints = new Vector<Matrix>();	
		// extra weights to reduce the number of high-bin points
		Vector<Double> extraWeights = new Vector<Double>();

		// number of points you need to try out, to get montePoints integration points
		int trialPoints = 0;
		
		int[] pBin = new int[bins]; // integration points in a bin
		double[] pureGauss = new double[bins]; // pure Gaussian likelihood
		double[] actual = new double[bins]; // actual likelihood
		for (int b = 0; b < bins; b++) 
		{
			pBin[b] = 0;
			pureGauss[b] = 0.0;
			actual[b] = 0.0;
		}
		
		// generate integration points
		double gMean = 0.0;
		double gMax = 0.5*standardDevs*standardDevs; // 5 S.D gives gMax = 12.5
		double sumWeights = 0.0;
		while (monteCarloPoints.size() < montePoints)
		{
			// generate a displacement to a random point in the tilted 3-D rectangle
			Matrix disp = corner.copy();
			// start at the corner and move back towards the centre along each side - maybe overshooting
			double rad2 = 0.0;
			for (int ev = 0; ev < 3; ev++)
			{
				double prop = 2.0* Math.random();
				rad2 = rad2 + (prop - 1.0)*(prop - 1.0); // maximum 1.0
				disp = disp.plus(side[ev].times(prop));
			}
			
			// calculate Gaussian cutoff for displacements; standardDevs  = 5 standard deviations
			// increase in negative log likelihood in the Gaussian approximation
			double g = disp.transpose().times(deriv2.times(disp)).get(0, 0);

			// Util.message(rad2 + "\t" + g + "\t" + g/rad2 + "\t" + gMax);
			
			// keep the point if it is within scaler standard deviations of the maximum likelihood point
			if (g < gMax) 
			{
				
				double v = Math.sqrt(g/gMax);
				int b = (int)(dBins*Math.sqrt(g/gMax));
				
				// points have weight 1 in lower bins
				if (v < vMin)
				{
					monteCarloPoints.add(disp);
					extraWeights.add(Double.valueOf(1.0));
					pBin[b]++;
					sumWeights = sumWeights + 1.0;
					gMean = gMean + g;
				}
				
				// fewer points in high bins, with larger weight
				else if (v > vMin)
				{
					double p = (vMin/v)*(vMin/v);
					if (Math.random() < p)
					{
						monteCarloPoints.add(disp);
						extraWeights.add(Double.valueOf(1.0/p));						
						pBin[b]++;
						sumWeights = sumWeights + 1.0/p;
						gMean = gMean + g;
					}
				}				
				
			}
			trialPoints++;
			if (trialPoints > 10*montePoints)
			{
				montePoints = 0; // to stop the loop
				throw new Exception("Failed to make Monte Carlo points");
			}
		}
		gMean = gMean/sumWeights;

		// check messages
		if (write) 
		{
			Util.message("\n" + montePoints + " integration points made from " + trialPoints + " trials; mean g = " + Util.round(gMean, 4) + "/" +gMax);
			Util.message("Sum of weights: " + sumWeights);
		}
				
		// initialise zeroth, first and second moments of actual likelihood to be integrated
		double zeroMoment = 0.0;
		Matrix firstMoment = Matrix.identity(3, 1).times(0.0);
		Matrix secondMoment = Matrix.identity(3, 3).times(0.0);
		
		// initialise checks by re-computing theoretical  Gaussian likelihood distribution
		double zeroMomentGaussCheck = 0.0;
		Matrix firstGaussCheck =  Matrix.identity(3, 1).times(0.0);
		Matrix secondGaussCheck =  Matrix.identity(3, 3).times(0.0);
		

		// loop over integration points
		for (int p = 0; p < montePoints; p++)
		{
			// the weight of each point is divided by the sum of all weights,
			// with extra weight in high bins, because the number of points in high bins has been reduced
			double weight = extraWeights.get(p).doubleValue()/sumWeights;
			
			// calculate true negative log likelihood at the point
			Matrix disp = monteCarloPoints.get(p);
			Matrix point = bestEstimate.plus(disp);
			double gap = negLogLikelihood(objx, Util.makeVec(point),null, calcMethod) - centreNegLog;// expected to be positive
			double trueLik = Math.exp(-gap);
			
			// integrate ideal Gaussian negative log likelihood for check
			double g = disp.transpose().times(deriv2.times(disp)).get(0, 0);
			double gaussLik = Math.exp(-g);
			
			// bin for histograms
			int b = (int)(dBins*Math.sqrt(g/gMax));
			
			// increment the set of likely points (within one standard deviation of the maximum real likelihood)
			if (gap < 0.5) likelyPoints.add(point);
			
			// increment moments
			double contribution = weight*trueLik;
			zeroMoment = zeroMoment + contribution;
			actual[b] = actual[b] + contribution;
			firstMoment = firstMoment.plus(disp.times(contribution));
			secondMoment = secondMoment.plus(disp.times(disp.transpose().times(contribution)));
			
			// check of ideal Gaussian likelihood distribution
			double gaussContribution  = weight*gaussLik;
			zeroMomentGaussCheck = zeroMomentGaussCheck + gaussContribution;
			pureGauss[b] = pureGauss[b] + gaussContribution;
			firstGaussCheck = firstGaussCheck.plus(disp.times(gaussContribution));
			secondGaussCheck = secondGaussCheck.plus(disp.times(disp.transpose().times(gaussContribution)));
		}
		
		double ratio = Util.round(zeroMomentGaussCheck/trueVolume,6);
		if (write)
		{
			Util.message("Theoretical Gaussian: " + trueVolume + "\tIntegrated Gaussian: " + zeroMomentGaussCheck 
					+ "\t(ratio " + ratio +  ")\nReal zeroth moment); " + zeroMoment);

			writeHistograms(montePoints,standardDevs, pBin, pureGauss, actual);
					
			Util.message("Zeroth moment real/gaussian ratio: " + Util.round(zeroMoment/zeroMomentGaussCheck, 2));
			Util.message("Likely points: " + likelyPoints.size());
		}

		Matrix[] result = new Matrix[6];

		// real likelihood results
		result[0] = Matrix.identity(1, 1).times(zeroMoment);
		result[1] = firstMoment;
		result[2] = secondMoment;

		// Gaussian checks
		result[3] = Matrix.identity(1, 1).times(zeroMomentGaussCheck);
		result[4] = firstGaussCheck;
		result[5] = secondGaussCheck;
		
		return result;
		
	}

	// ---------------------------------------------------------------------------------------------------
	//                                Support for estimation	
	// ---------------------------------------------------------------------------------------------------

	/**
	 * 
	 * @param slope
	 * @param smallesEval
	 * @param trials
	 * @return
	 */
	private boolean useMatrix(double slope, double smallestEval, int trials)
	{
		boolean matrix = false;
		if ((smallestEval > 5.0))
		{
			if (slope < 10) matrix = true;
		}
		return matrix;
			
		
	}
	
	
	/**
	 * 
	 * @param obj
	 * @param estPosition
	 * @return
	 */
	private boolean stableStartPoint (int obj, Matrix estPosition, boolean isFinal, int trace, int calcMethod)
	{
		double maxEVal = largestEigenvalue(obj,estPosition,calcMethod);
		boolean stable = (maxEVal > 25.0);
		
		if (stable)
		{
			double slope = slope(obj,estPosition,calcMethod);
			String fin = "";
			if (isFinal) fin = "Final ";
			if (trace == 5) Util.message(fin + "Slope for object " + obj + " at minimum: " + slope );
			
		}
		
		return stable;
	}
	
	
	
	/**
	 * 
	 * @param obj
	 * @param estimate
	 * @return magnitude of the slope
	 */
	private double slope(int obj, Matrix estimate, int calcMethod)
	{
		Matrix vFirst = firstDeriv(obj, Util.makeVec(estimate),calcMethod);			
		double slope = Math.sqrt(vFirst.transpose().times(vFirst).get(0, 0));
		return slope;		
	}

	/**
	 * 
	 * @param obj
	 * @param estimate
	 * @return the smallest eigenvalue of the second derivative matrix
	 */
	double smallestEigenvalue (int obj, Matrix estimate, int calcMethod)
	{
		Matrix second = secondDeriv(obj, Util.makeVec(estimate),calcMethod);
		return smallestEigenvalue(second);
	}
	
	/**
	 * 
	 * @param second
	 * @return
	 */
	double smallestEigenvalue (Matrix second)
	{
		EigenvalueDecomposition eComp = second.eig();
		Matrix eVals = eComp.getD();
		double smallest = 100000;
		
		for (int dim = 0; dim < 3; dim++)
		{
			double eVal = eVals.get(dim,dim);
			if (eVal < smallest) smallest= eVal;
		}
		
		return smallest;
	}
	
	

	/**
	 * 
	 * @param obj
	 * @param estimate
	 * @return the largest eigenvalue of the second derivative matrix
	 */
	double largestEigenvalue (int obj, Matrix estimate, int calcMethod)
	{
		Matrix second = secondDeriv(obj, Util.makeVec(estimate),calcMethod);
		EigenvalueDecomposition eComp = second.eig();
		Matrix eVals = eComp.getD();
		double largest = -100000;
		
		for (int dim = 0; dim < 3; dim++)
		{
			double eVal = eVals.get(dim,dim);
			if (eVal > largest) largest= eVal;
		}
		
		return largest;
	}
	

	private void debugTrace(String x) 
	{
		boolean debug = false;
		if (debug) Util.message(x);
	}


}
