package mPack;

import java.util.Collections;
import java.util.Vector;

import Jama.Matrix;
import utilPack.Util;

public class Motion {
	
	private Truth truth;

	private SenseData sense;
	
	private SFM_Model model;
	
	private Estimator estimator;
	
	int[] calcMethod = {Estimator.RUN_BAYES,Estimator.RUN_BAYES,Estimator.RUN_TRACK,Estimator.RUN_NOISY_TRACK};
	
	/* threshold gap in negative log likelihood, between the median of points thought to be in the moving shape, 
	 * and the median of other points. Do not test for motion if the gap is less than this. */
	private double likelyGapThreshold = 2.0;

	// improvement in negative log likelihood required to move all point thought to be in the shape
	double ImprovementThresholdForMove = 3.0;
	
	// true if a point is judged to be moving in a step, by an estimation method
	public boolean[][][] moving = new boolean [SFM_Model.maxObjects][Estimator.estTypes][SFM_Model.maxSteps];
	
	// sum over trials, of 1  if more than half the points in a shape are judged to be moving in a step, by an estimation method; 0 if half or fewer
	public int[][] sumMoving = new int [Estimator.estTypes][SFM_Model.maxSteps];
	
	// maximum number of different noise levels to be plotted
	public int noiseLevels = 5;
	
	// percentage success score, averaged over trials, at different noise levels;
	public int[][] percentSuccess = new int [SFM_Model.maxSteps][noiseLevels];
	
	// from class Estimator
	
	/* types of estimate
	public static int FULLBAYES = 0;
	public static int MEAN = 1;
	public static int TRACK = 2;
	public static int NOISYTRACK = 3;
	*/
		
	/* allowed values of calcMethod	
	public static int START_BAYES = 0;
	public static int RUN_BAYES = 1;
	public static int START_TRACK = 2;
	public static int RUN_TRACK = 3;
	public static int START_NOISY_TRACK = 4;
	public static int RUN_NOISY_TRACK = 5;
	*/
	
	// trials for a graph of success rate vs. steps
	
	// maximum number of trials
	public static int maxTrials = 60;
	
	// currently set number of trials
	public int trials = 30;
	
	// number of steps to be plotted
	public int trialSteps;
	
	
	
	//-----------------------------------------------------------------------
	//                      Constructor
	//-----------------------------------------------------------------------

	
	public Motion(Truth truth,SenseData sense,Estimator estimator)
	{
		this.truth = truth;
		this.sense = sense;
		this.estimator = estimator;
	}

	
	public void setModel(SFM_Model model) {this.model = model;}

	
	//-----------------------------------------------------------------------------------------
	//                 attempt to detect motion at each step
	//-----------------------------------------------------------------------------------------
	
	/**
	 * 
	 * @param step
	 */
	public void detectMotion(int step)
	{
		// number of vertices in the moving shape
		int verts = truth.allPoints - truth.randomPoints;
		
		// only detect motion if there is a moving shape
		if (verts > 0)
		{
			//loop over two estimation methods (TRACK and NOISY_TRACK)
			for (int estType = 0; estType < 4; estType++)
			{
				// miss out the FULLBAYES and MEAN estimation methods; only TRACK and NOISY_TRACK
				if (estType > Estimator.MEAN) // i.e only do estTye = 2 and 3
				{
					boolean trace = estType == 2;
					// collect the negative log likelihoods at the fitted points
					Vector<LikelyOrder> points = new Vector<LikelyOrder>();
					int pointsEstimated = 0;
					for (int pobj = 0; pobj < truth.allPoints; pobj++) if (estimator.estimated[estType][step][pobj])
					{							
						double logLik = estimator.negLogLikelihood[estType][step][pobj];
						Matrix pos = estimator.locationEstimate[estType][step][pobj];
						LikelyOrder lo = new LikelyOrder(pobj,step,pos,estType,logLik);
						points.add(lo);
						pointsEstimated++;
					}
					
					// minimum number of points estimated  for motion detection
					if (pointsEstimated > verts + 4)
					{
						Collections.sort(points);
						String summary = "\nStep " + step + ",estimation type " + estType + ": ";
						for (LikelyOrder lo: points) summary = summary + "(" + lo.object + ") " + Util.round(lo.likely, 4) + ", ";
						// Util.message(summary);
						
						// find the  median negative log likelihood of the first vert points, and of the remaining points
						int inMedian = verts/2;
						double inLikely = points.get(inMedian).likely;
						int outMedian = verts + (pointsEstimated - verts)/2;
						double outLikely = points.get(outMedian).likely;
						// Util.message("Median negative log likelihoods: " + Util.round(inLikely, 4) + "\t" + + Util.round(outLikely, 4) + "\n");
						double likelyGap = inLikely - outLikely ;
						
						// there is enough likelihood gap between the median points of in and out groups
						if (likelyGap > likelyGapThreshold)
						{
							
							markOutliers(verts, points);
							
							// only 0, 1 or 2 outliers
							int outliers = countOutliers(points);
							if (outliers < 3) tryMovingPoints(verts,points,estType,step,trace);
						}
					}
					
				}
			}
		}
	}
	
	
	/**
	 * try moving points in the in group all by the same (x,y,z) to reduce their 
	 * aggregate negative log likelihood
	 * 
	 * @param verts number of points in the in group 
	 * @param points Vector of all points and their negative log likelihoods (some may be outliers or failed)
	 */
	private void tryMovingPoints(int verts,Vector<LikelyOrder> points, int estType, int step1, boolean trace)	
	{
		if(trace) Util.message("\nChecking for movement in step " + step1);
		
		Vector<LikelyOrder> movingPoints = new Vector<LikelyOrder>();
		for (int v = 0; v < verts; v++) movingPoints.add(points.get(v));
		
		// calculate  initial aggregate negative log likelihood of the points, with no movement
		Matrix disp = Matrix.identity(3, 1).times(0.0);
		double firstNegLogLik = likelySum(movingPoints, disp);

		// prepare for iterative search
		double negLogLik = 1000.0;		
		boolean going = true;
		double movementStep = 0.005;
		double previousNLL = 10000.0;
		double epsDeriv = 0.0001;
		int iteration = 0;
		int maxSteps = 100;

		// gradient descent search for best aggregate negative log likelihood, moving all points together
		for (int iter = 0; iter <maxSteps; iter++) if (going)
		{
			iteration = iter;

			negLogLik = likelySum(movingPoints, disp);
			// double dispMod = Math.sqrt(disp.transpose().times(disp).get(0, 0));

			
			// gradient vector of aggregate negative log likelihood
			Matrix slope = likelySum1D(movingPoints, disp,epsDeriv);
			double slopeMod = Math.sqrt(slope.transpose().times(slope).get(0, 0)); // modulus of slope
			Matrix delta = slope.times(movementStep/slopeMod);	// recommended step in (x,y,z)		
			disp.minusEquals(delta);

			// Util.message("NLL: " + negLogLik + "\tSlope: " + slopeMod + "\tDisp: " + dispMod);

			// require continuous improvement to keep going
			going = (negLogLik < previousNLL);
			previousNLL = negLogLik;
		}
		
		double improvement = firstNegLogLik - negLogLik;

		if (trace)
		{
			double dist = Math.sqrt(disp.transpose().times(disp).get(0, 0));
			Util.message("\nMove by " + Util.round(dist,4) + " after " + iteration + " iterations with improvement " + improvement + "\t"
					+ Util.round(disp.get(0, 0), 4) + "\t"  + Util.round(disp.get(1, 0), 4) + "\t" + Util.round(disp.get(2, 0), 4));
		}
		
		// if there is enough improvement, move all points whose movement has been found
		if (improvement  > ImprovementThresholdForMove)
		{
			if (trace) Util.message("\nMoving points for estimation type " + estType);
			String moved = "";
			int inShapeMoved = 0; // the number of points in the shape found to move
			for (LikelyOrder lo: movingPoints) if (lo.usable())
			{
				Matrix startPos =  estimator.locationEstimate[estType][lo.step][lo.object].copy();
				Matrix endPos = startPos.plus(disp);
				estimator.locationEstimate[estType][lo.step][lo.object] = endPos.copy();
				
				if (trace) Util.message("Moved object " + lo.object + "\tfrom " + Util.mString(startPos) + "\tto " + Util.mString(endPos));
				
				moving[lo.object][estType][lo.step] = true;				
				moved = moved + Integer.valueOf(lo.object).toString() + " ";

				// points in the shape, not static
				if (lo.object > truth.randomPoints - 1) inShapeMoved++;
			}
			if (inShapeMoved > verts/2) sumMoving[estType][step1]++;
			if (trace) Util.message("\nMoved points: " + moved + "; \t" + inShapeMoved + " points in the shape.\n");
		}
	}


	/**
	 * 
	 * @param points
	 * @param displacement
	 * @return likelihood sum for a set of points
	 */
	private double likelySum(Vector<LikelyOrder> points, Matrix displacement)
	{
		double likSum = 0.0;
		// Util.message("Likely sum displacement): " + displacement.get(0, 0) + "\t"+ displacement.get(1, 0) + "\t"+ displacement.get(2, 0));
		
		for (LikelyOrder lo: points) if (lo.usable()) 
		{
			double part = lo.dispLikely(displacement);
			likSum = likSum + part;
			// Util.message("object " + lo.object + "; " + part);			
		}
		
		return likSum;
	}
	
	/**
	 * 
	 * @param points
	 * @param displacement
	 * @return first derivative of likelihood sum for a set of points
	 */
	private Matrix likelySum1D(Vector<LikelyOrder> points, Matrix displacement,double eps)
	{
		Matrix deriv1 = Matrix.identity(3, 1).times(0.0);
		
		for (int dim = 0; dim < 3; dim++)
		{
			Matrix dPlus = displacement.copy();
			dPlus.set(dim, 0, dPlus.get(dim, 0) + 0.5*eps);
			Matrix dMinus = displacement.copy();
			dMinus.set(dim, 0, dMinus.get(dim, 0) - 0.5*eps);
			
			double d1 = likelySum(points,dPlus);
			double d2 = likelySum(points,dMinus);
			
			double diff = d1 - d2;
			deriv1.set(dim,0,diff/eps);
			
			// Util.message("1st deriv, dim = " + dim + ":\t" + d1 + "\t" + d2 + "\t" + diff);
		}
		
		return deriv1;
	}
	
	/**
	 * 
	 * @param points
	 * @param displacement
	 * @return second derivative of likelihood sum for a set of points
	 */
	private Matrix likelySum2D(Vector<LikelyOrder> points, Matrix displacement)
	{
		double eps = 0.0001;
		Matrix deriv2 = Matrix.identity(3, 3).times(0.0);
		
		// the dimension being varied by eps
		for (int dim = 0; dim < 3; dim++)
		{
			Matrix dPlus = displacement.copy();
			dPlus.set(dim, 0, dPlus.get(dim, 0) + 0.5*eps);
			Matrix dMinus = displacement.copy();
			dMinus.set(dim, 0, dMinus.get(dim, 0) - 0.5*eps);
			
			// matrix/vector  of differences
			Matrix diff = likelySum1D(points,dPlus,eps).minus(likelySum1D(points,dMinus,eps));
			
			// transcribe the difference components in the other dimension
			for (int d = 0; d < 3; d++)
			{
				deriv2.set(dim,d,diff.get(d, 0)/eps);
			}
			
		}
		
		return deriv2;
	}
	

	/**
	 * mark the points which are outliers:
	 * whose distance from the centre of gravity 
	 * is more than three times  the mean for the group
	 * @param verts
	 * @param points
	 */
	private void markOutliers(int verts,Vector<LikelyOrder> points)
	{
		// find the centre of gravity of the first verts points
		Matrix cOfG = Matrix.identity(3, 1).times(0.0);
		for (int p = 0; p < verts;p++) cOfG.plusEquals(points.get(p).pos.times(1.0/verts));
		
		// find the mean distance from the centre of gravity
		double meanDist = 0.0;
		for (int p = 0; p < verts;p++) 
		{
			Matrix disp = cOfG.minus(points.get(p).pos);
			double dist = Math.sqrt(disp.transpose().times(disp).get(0, 0));
			meanDist = meanDist + dist/verts;
		}
		
		//mark the outliers
		for (int p = 0; p < verts;p++) 
		{
			Matrix disp = cOfG.minus(points.get(p).pos);
			double dist = Math.sqrt(disp.transpose().times(disp).get(0, 0));
			if (dist > 3.0*meanDist) points.get(p).outlier = true;
		}
	}
	
	/**
	 * 
	 * @param points
	 * @return the number of outliers
	 */
	private int countOutliers(Vector<LikelyOrder> points)
	{
		int outliers = 0;
		
		for (LikelyOrder point: points) if (point.outlier) outliers++;
		
		return outliers;
	}

	
	//-----------------------------------------------------------------------------------------------
	//                  Inner class for sorting negative log likelihoods of objects
	//-----------------------------------------------------------------------------------------------
	
	class LikelyOrder implements Comparable<LikelyOrder>
	{
		int object;
		double likely;
		double prevLikely;
		boolean outlier = false;
		Matrix pos;
		int estType;
		int step;
		
		boolean failed = false;
		
		boolean usable() {return (!outlier && !failed);}
		
		/**
		 * 
		 * @param object
		 * @param step
		 * @param pos
		 * @param estType
		 * @param likely
		 */
		LikelyOrder(int object, int step, Matrix pos, int estType, double prevLikely)
		{
			this.object = object;
			this.step = step;
			this.pos = pos;
			this.estType = estType;
			this.prevLikely = prevLikely;
			
			likely = dispLikely(null);
								
			if (estType == Estimator.TRACK) 
			{
				checkZeroDisplacement();
				// Util.message("Checked: " + !failed);
			}
			
			// Util.message("Made LikelyOrder for object " + object + ",step " + step + ", estType " + estType + "; failed = " + failed);
		}
		
		public int compareTo(LikelyOrder other)
		{
			if (other.likely > likely) return 1;
			return -1;
		}
		
		/**
		 * 
		 * @param displacement
		 * @return negative log likelihood with some displacement of the previous best estimate
		 */
		public double dispLikely(Matrix displacement)
		{
			double dLik = 0.0;
			try
			{
				double[] posVec = Util.makeVec(pos);
				dLik = estimator.negLogLikelihood(object, posVec, displacement, calcMethod[estType]);
			}
			catch (Exception ex) 
			{
				 /* Util.message("Failed to re-compute negative log likelihood for object " 
							+ object + ", estType " + estType + ", step " + step); */
				failed = true;
			}
			return dLik;
		}
	
		/**
		 * check that with zero displacement, you get the same negative log likelihood as has been stored
		 * this check should work for TRACK, not NOISYTRACK
		 */
		void checkZeroDisplacement()
		{
			if (!failed)
			{
				double diff = prevLikely - likely;
				if (Math.abs(diff) > 0.01) 
				{
					Util.message("Cannot reproduce stored NLL: " + Util.round(prevLikely,4) + "\t" + Util.round(likely ,4)
							+ " for object " + object  + " step " + step);
					double repeat = dispLikely(null);
				}
				
			}
		}
		
	}
	
	
    //--------------------------------------------------------------------------------------
    //                               Motion Detection trials 
    //--------------------------------------------------------------------------------------
	
	/**
	 * set everything not ranked
	 */
	public void startMotionCounts(boolean zeroCount)
	{
		for (int s = 0; s < model.maxSteps; s++)
		{
			for (int m = 0; m < Estimator.estTypes; m++)
			{
				for (int p =0; p < model.maxObjects; p++) moving[p][m][s] = false;					
				if (zeroCount) sumMoving[m][s] = 0;					
			}
			for (int nl = 0; nl < noiseLevels; nl++) percentSuccess[s][nl] = 0;
		}

	}


	
	/**
	 * compare efficiency of detecting motion, for tracking and tracking with memory noise
	 */
	public void detectMotion()
	{
		// if there is a shape to move
		if (truth.isShape)
		{
			// record which shape is being moved
			int shapeUsed = truth.shapeSelected;

			// set counts over trials to zero,  for all objects for all steps of all trials
			startMotionCounts(true);

			// set up the number of steps per trial
			trialSteps = (int)(truth.startStep + 2*truth.inSteps);
			if (trialSteps > SFM_Model.maxSteps) trialSteps = SFM_Model.maxSteps;
			
			// loop over trials
			for (int trial = 0; trial < trials; trial++)
			{
				// start again with new random object positions and animal track
				model.start(false);
				
				// re-introduce the shape at a random position
				truth.placeShape(shapeUsed);
				
				// loop over steps
				for (int step = 0; step < trialSteps; step++)
				{
					model.step(false);

					// doing a step calls detectMotion(step); no need to call it from here
				}
			}
			
			// compute success scores at two different noise levels
			computeSuccessScores();
			
			// display a graph of results in the right hand  window
			model.setCaption(2, SFM_Model.MOTION);
			model.doDrawingAndCaption(2);
			
		}
		
	}

	/**
	 * compute the level of success in detecting motion, 
	 * as f8nction of steps, for tracking and noisy tracking
	 */
	private void computeSuccessScores()
	{
		Util.message("\nMotion Detection scores");
		int verts = truth.allPoints - truth.randomPoints;

		// the only two graphs so far
		int[] plotTypes = {Estimator.TRACK, Estimator.NOISYTRACK};
		
			for (int step = 0; step < trialSteps;step++)
			{
				for (int nl = 0; nl < 2; nl++)
				{
					int estType = plotTypes[nl];
					
					// count 1 for each trial in which  the shape is found to be moving; maximum possible = trials
					percentSuccess[step][nl] = (100* sumMoving[estType][step])/trials;
				}
				
				Util.message("Step " + step + " scores: " + percentSuccess[step][0] + "\t" + percentSuccess[step][1]);
			}
				
	}

}
