package mPack;

import java.util.Vector;

import Jama.Matrix;
import display.DrawArea;
import utilPack.Util;

public class StepData {
	
	private SFM_Model model;
	
	private Truth truth;
	
	private SenseData sense;
	
	private Estimator estimator;
	
	private int step; 
	
	private int species;
	
	private int maxObjects = SFM_Model.maxObjects;
	
	private int nObjects;
	
	public boolean hasData;
	
	
    //--------------------------------------------------------------------------------------
    //              Dynamical model data - ground truth
    //--------------------------------------------------------------------------------------
	
	
	// animal's (x,y,z) velocity
	public double[] animalVelocity = new double[3];	
	
	// animal's position 
	public double[] animalPosition = new double[3];	
	
    //--------------------------------------------------------------------------------------
    //              Dynamical model data - animal's sense data
    //--------------------------------------------------------------------------------------
	
		
	// visual data of the bee; for each object; a unit vector  in [x,y,z]
	public double[][] beeVisionData = new double[maxObjects][3];
	
	// echo-location data for each object
	// the last double [5] is [ux, uy, uz, range, doppler shift(cos theta) ] 
	// where u is the unit vector along the  bat's velocity
	public double[][] batSonarData = new double[maxObjects][5];

	
    //--------------------------------------------------------------------------------------
    //              Dynamical model data - animal's estimates 
    //--------------------------------------------------------------------------------------
	
	static int estTypes = 6;
	
	
	// true if an object has been placed in any step (do not draw the estimated object location if this is false)
	public boolean[][] estimated = new boolean[estTypes][maxObjects];
	
	// maximum likelihood locations of objects estimated at each step, 3*1 Matrix
	public Matrix[][] locationEstimate = new Matrix[estTypes][maxObjects];
			
	// second derivative of negative log likelihood
	public Matrix[][] negLog2Deriv = new Matrix[estTypes][maxObjects];

	// 3*3 matrix of error bars around maximum likelihood locations. Each column (second index) = one orthogonal error bar, with length = one standard deviation
	public Matrix[][] locErrorBars = new Matrix[estTypes][maxObjects];

	
	//-----------------------------------------------------------------------------
	//
	//-----------------------------------------------------------------------------
	
	/**
	 * 
	 */
	public StepData()
	{
		hasData = false;
	}
	
	/**
	 * 
	 * @param model
	 * @param species
	 * @param step
	 */
	public StepData(SFM_Model model, Truth truth, SenseData sense, Estimator estimator, int species, int step)
	{
		this.model = model;
		this.truth = truth;
		this.sense = sense;
		this.estimator = estimator;
		this.step = step;
		this.species = species;
		
		hasData = true;
		
		nObjects = (int)truth.allPoints;
		
		for (int d = 0; d < 3; d++)
		{
			animalVelocity[d]  = truth.animalVelocity[step][d];
			animalPosition[d]  = truth.animalPosition[step][d];
		}
				

		for (int obj = 0; obj < nObjects; obj++)
		{
			for (int d = 0; d < 3; d++)
				beeVisionData[obj][d] = sense.beeVisionData[step][obj][d];
			
			for (int dp = 0; dp < 5; dp++)
				batSonarData[obj][dp] = sense.batSonarData[step][obj][dp];
			
			for (int type = 0; type < 4;type++)
			{
				estimated[type][obj] = estimator.estimated[type][step][obj];
				
				if (estimated[type][obj])
				{
					locationEstimate[type][obj] = estimator.locationEstimate[type][step][obj];
					negLog2Deriv[type][obj] = estimator.negLog2Deriv[type][step][obj];
					locErrorBars[type][obj] = estimator.locErrorBars[type][step][obj];
				}
			}

		}
	}

	/**
	 * cloning constructor
	 * @param sd
	 */
	public StepData(StepData sd)
	{
		this.model = sd.model;
		this.step = sd.step;
		this.species = sd.species;
		
		hasData = sd.hasData;
		
		nObjects = (int)truth.allPoints;
		
		if (hasData) for (int d = 0; d < 3; d++)
		{
			animalVelocity[d]  = sd.animalVelocity[d];
			animalPosition[d]  = sd.animalPosition[d];
		}
		
		if (hasData) for (int obj = 0; obj < nObjects; obj++)
		{
			for (int d = 0; d < 3; d++)
				beeVisionData[obj][d] = sd.beeVisionData[obj][d];
			
			for (int dp = 0; dp < 5; dp++)
				batSonarData[obj][dp] = sd.batSonarData[obj][dp];
			
			for (int type = 0; type < 4;type++)
			{
				estimated[type][obj] = sd.estimated[type][obj];
				
				if (estimated[type][obj])
				{
					locationEstimate[type][obj] = sd.locationEstimate[type][obj].copy();
					negLog2Deriv[type][obj] = sd.negLog2Deriv[type][obj].copy();
					locErrorBars[type][obj] = sd.locErrorBars[type][obj].copy();
				}
			}

			
		}
	}
	
	//-----------------------------------------------------------------------------
	//
	//-----------------------------------------------------------------------------
	
	/**
	 * 
	 * @param write
	 * @return
	 */
	public double[] writeMeanErrorCompare(boolean write)
	{
		// do not add proportional errors for an object to the sum over objects, if (tracked proportional error squared > threshold squared)
		double threshold = 0.2;
		
		// for both tracked and noisy tracked, divide by sqrt(range**2 + minRange**2), to damp the effect of very close objects
		double minRange = 0.3;
		
		if (write)
		{
			Util.message("\nMean Error Comparison at step " + step);
			Util.message("\nObj\tColour\tR\t\tE2T\tE2N\n");
		}
				
		double trackedPropErr = 0.0;
		double noisyPropErr = 0.0;
		int trackedCount = 0;
		int noisyCount = 0;
		for (int  obj = 0; obj < nObjects; obj++)
		{
			Matrix  realObj = truth.objectLocation[step][obj];
			Matrix animalPos = Util.makeMatrix(animalPosition);
			Matrix realDiff = realObj.minus(animalPos);

			// squared length of the real difference (dimension d**2)
			double norm2 = dotP(realDiff, realDiff);
			// denominator for ratio of error/range (so range cannot be less than minRange)
			double norm = Math.sqrt(norm2 + minRange*minRange);
			
			// part of the row that can always be written
			String lr = (obj + "\t" + DrawArea.colour(obj) + "\t" + Util.round(norm, 4)) + "\t\t" ;
			String lt = "-\t";
			String ln = "-";
			
			// part that can be written if the object is tracked, and if tracking has not gone off the rails
			if (estimated[estimator.TRACK][obj])
			{
				double propDistError = propError(obj,false);
				// only filter on tracked proportional error, not noisy proportional error
				if (propDistError < threshold)
				{
					trackedCount++;
					trackedPropErr = trackedPropErr + propDistError;
					lt = Util.round(trackedPropErr,4) + "\t";

					// part that can be written if the object is tracked with memory noise
					if (estimated[estimator.NOISYTRACK][obj])
					{
						double noisyPropDistError = propError(obj,true);
						noisyCount++;
						noisyPropErr = noisyPropErr + noisyPropDistError;
						ln = Util.round(noisyPropErr,4)  + " ";
					}
				}
			}
					
			
			if (write) Util.message(lr + lt  + ln);
			
		}
		
		// summaries over all objects
		double[] result = {0.0,0.0};

		double trackedMean = 0.0;
		if (trackedCount > 0)
		{
			trackedMean = trackedPropErr/trackedCount;
			result[0] = trackedMean;
		}

		double noisyMean = 0.0;
		if (noisyCount > 0) 
		{
			noisyMean = noisyPropErr/noisyCount;
			result[1] = noisyMean;
		}
		
		if (write)
		{
			Util.message("Mean\t\t\t\t" + Util.round(trackedMean, 4) + "\t" + Util.round(noisyMean, 4));
		}
		
		
		return result;
	}

	
	/**
	 * 
	 * @param write
	 * @return
	 */
	public double[] writeRMSErrorCompare(boolean write)
	{
		// do not add proportional errors for an object to the sum over objects, if (tracked proportional error squared > threshold squared)
		double threshold = 0.2;
		
		if (write)
		{
			Util.message("\nRMS Error Comparison at step " + step);
			Util.message("\nObj\tColour\tR\t\tE2T\tE2N\n");
		}
				
		double trackedRangeSumSq = 0.0;
		double noisyRangeSumSq = 0.0;
		int trackedCount = 0;
		int noisyCount = 0;
		for (int  obj = 0; obj < nObjects; obj++)
		{
			Matrix  realObj = truth.objectLocation[step][obj];
			Matrix animalPos = Util.makeMatrix(animalPosition);
			Matrix realDiff = realObj.minus(animalPos);

			// squared length of the real difference (dimension d**2)
			double norm2 = dotP(realDiff, realDiff);
			// length of the real difference (dimension d)
			double norm = Math.sqrt(norm2);
			
			// part of the row that can always be written
			String lr = (obj + "\t" + DrawArea.colour(obj) + "\t" + Util.round(norm, 4)) + "\t\t" ;
			String lt = "-\t";
			String ln = "-";
			
			// part that can be written if the object is tracked, and if tracking has not gone off the rails
			if (estimated[estimator.TRACK][obj])
			{
				double propRangeErrorSq = rangeErrorSquared(obj,false);
				// only filter on tracked proportional error, not noisy proportional error
				if (propRangeErrorSq < threshold*threshold)
				{
					trackedCount++;
					trackedRangeSumSq = trackedRangeSumSq + propRangeErrorSq;
					lt = Util.round(propRangeErrorSq,4) + "\t";

					// part that can be written if the object is tracked with memory noise
					if (estimated[Estimator.NOISYTRACK][obj])
					{
						double noiseRangeErrorSq = rangeErrorSquared(obj,true);
						double noisyPropDistError = propError(obj,true);
						noisyCount++;
						noisyRangeSumSq = noisyRangeSumSq + noiseRangeErrorSq;
						ln = Util.round(noiseRangeErrorSq,4)  + " ";
					}
				}
			}
					
			
			if (write) Util.message(lr + lt  + ln);
			
		}
		
		// summaries over all objects
		double[] result = {0.0,0.0};

		double trackedAvgSQ = 0.0;
		double trackedRMS = 0.0;
		double rmsRatio = 0.0;
		if (trackedCount > 0)
		{
			trackedAvgSQ = trackedRangeSumSq/trackedCount;
			trackedRMS = Math.sqrt(trackedAvgSQ);
			result[0] = trackedRMS;
		}

		double noisyAvgSq = 0.0;
		double noisyRMS = 0.0;
		if (noisyCount > 0) 
		{
			noisyAvgSq = noisyRangeSumSq/noisyCount;
			noisyRMS = Math.sqrt(noisyAvgSq);
			result[1] = noisyRMS;
			rmsRatio = noisyRMS/trackedRMS;
		}
		
		if (write)
		{
			Util.message("Mean\t\t\t\t" + Util.round(trackedAvgSQ, 4) + "\t" + Util.round(noisyAvgSq, 4));
			Util.message("RMS\t\t\t\t" + Util.round(trackedRMS, 4) + "\t" + Util.round(noisyRMS, 4) + "\t" + Util.round(rmsRatio, 4));
		}
		
		
		return result;
	}
	
	
	//-----------------------------------------------------------------------------
	//
	//-----------------------------------------------------------------------------
	
	/**
	 * 
	 * @param obj
	 * @param noisy
	 * @return the error in estimating the object position,
	 * divided by the true distance from the animal to the object
	 */
	public double propError(int obj, boolean noisy)
	{
		// true distance from the animal to the object
		Matrix  realObj = truth.objectLocation[step][obj];
		Matrix animalPos = Util.makeMatrix(animalPosition);
		Matrix realDiff = realObj.minus(animalPos);
		double dist = Math.sqrt(realDiff.transpose().times(realDiff).get(0, 0));
		
		// estimated object position, with and without noise
		Matrix estObj = locationEstimate[estimator.TRACK][obj];
		if (noisy) estObj = locationEstimate[estimator.NOISYTRACK][obj];
		
		// distance from the estimate to the true object position
		Matrix estDiff = estObj.minus(realObj);
		double eDist = Math.sqrt(estDiff.transpose().times(estDiff).get(0, 0));
		
		return eDist/dist;
	}
	
	/**
	 * 
	 * @param obj
	 * @param noisy
	 * @return squared (range error/true range)
	 */
	public double rangeErrorSquared(int obj, boolean noisy)
	{
		Matrix  realObj = truth.objectLocation[step][obj];
		Matrix animalPos = Util.makeMatrix(animalPosition);
		Matrix realDiff = realObj.minus(animalPos);
		
		Matrix estObj = locationEstimate[estimator.TRACK][obj];
		if (noisy) estObj = locationEstimate[estimator.NOISYTRACK][obj];
		Matrix estDiff = estObj.minus(animalPos);
		
		
		return propErrorSquared(realDiff,estDiff);
	}
	
	/**
	 * 
	 * @param realDif real displacement from the animal to the object
	 * @param estDiff estimated displacement from the animal to the object
	 * @return squared proportional error in the component of the estimate along the real displacement
	 */
	private double propErrorSquared(Matrix realDiff,Matrix estDiff)
	{
		// squared length of the real difference (dimension d**2)
		double norm2 = dotP(realDiff, realDiff);
		// length of the real difference (dimension d)
		double norm = Math.sqrt(norm2);
		
		// component of the estimate along the real vector (should be close to norm) (dimension d)
		double dot = dotP(realDiff, estDiff)/norm;
		
		// proportional error in the component of the estimate (dimension 1)
		double propErr = (dot - norm)/norm;
		
		// proportional error squared
		return propErr*propErr;
		
	}
	
	/**
	 * 
	 * @param a
	 * @param b
	 * @return
	 */
	private double dotP(Matrix a, Matrix b)
	{
		return a.transpose().times(b).get(0, 0);
	}

	

}
